# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path
import shutil
import subprocess
import sys
from textwrap import dedent

import colors
import pytest

from pex import dist_metadata
from pex.cli.commands.venv import InstallLayout
from pex.cli.testing import run_pex3
from pex.common import open_zip, safe_open
from pex.compatibility import commonpath
from pex.interpreter import PythonInterpreter
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pex import PEX
from pex.platforms import Platform
from pex.testing import IS_MAC, PY39, PY310, ensure_python_interpreter, make_env, run_pex_command
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import Virtualenv

if TYPE_CHECKING:
    from typing import Any


@pytest.fixture(scope="module")
def td(tmpdir_factory):
    # type: (Any) -> Any

    return tmpdir_factory.mktemp("td")


@pytest.fixture(scope="module")
def lock(td):
    # type: (Any) -> str

    lock = str(td.join("lock.json"))
    run_pex3(
        "lock", "create", "cowsay==5.0", "ansicolors==1.1.8", "-o", lock, "--indent", "2"
    ).assert_success()
    return lock


@pytest.fixture(scope="module")
def cowsay_pex(
    td,  # type: Any
    lock,  # type: str
):
    # type: (...) -> str

    pex = str(td.join("pex"))
    run_pex_command(args=["--lock", lock, "-o", pex]).assert_success()
    assert sorted(
        [(ProjectName("cowsay"), Version("5.0")), (ProjectName("ansicolors"), Version("1.1.8"))]
    ) == [(dist.metadata.project_name, dist.metadata.version) for dist in PEX(pex).resolve()]
    return pex


def test_venv_empty(tmpdir):
    # type: (Any) -> None

    dest = os.path.join(str(tmpdir), "dest")
    run_pex3("venv", "create", "-d", dest).assert_success()
    venv = Virtualenv(dest)
    assert (
        PythonInterpreter.get().resolve_base_interpreter()
        == venv.interpreter.resolve_base_interpreter()
    )
    assert [] == list(venv.iter_distributions())


def assert_venv(
    tmpdir,  # type: Any
    *extra_args  # type: str
):
    # type: (...) -> None

    dest = os.path.join(str(tmpdir), "dest")
    run_pex3("venv", "create", "cowsay==5.0", "-d", dest, *extra_args).assert_success()

    venv = Virtualenv(dest)
    _, stdout, _ = venv.interpreter.execute(
        args=["-c", "import cowsay, os; print(os.path.realpath(cowsay.__file__))"]
    )
    assert venv.site_packages_dir == commonpath([venv.site_packages_dir, stdout.strip()])

    _, stdout, _ = venv.interpreter.execute(args=["-m", "cowsay", "--version"])
    assert "5.0" == stdout.strip()

    assert (
        "5.0"
        == subprocess.check_output(args=[venv.bin_path("cowsay"), "--version"])
        .decode("utf-8")
        .strip()
    )

    assert [(ProjectName("cowsay"), Version("5.0"))] == [
        (dist.metadata.project_name, dist.metadata.version) for dist in venv.iter_distributions()
    ]


def test_venv(
    tmpdir,  # type: Any
    lock,  # type: str
    cowsay_pex,  # type: str
):
    # type: (...) -> None

    assert_venv(tmpdir)
    assert_venv(tmpdir, "--lock", lock)
    assert_venv(tmpdir, "--pex-repository", cowsay_pex)


def test_flat_empty(tmpdir):
    # type: (Any) -> None

    dest = os.path.join(str(tmpdir), "dest")
    run_pex3("venv", "create", "--layout", "flat", "-d", dest).assert_success()
    assert [] == list(dist_metadata.find_distributions(search_path=[dest]))


def test_flat_zipped_empty(tmpdir):
    # type: (Any) -> None

    dest = os.path.join(str(tmpdir), "dest")
    run_pex3("venv", "create", "--layout", "flat-zipped", "-d", dest).assert_success()
    assert [] == list(dist_metadata.find_distributions(search_path=[dest]))
    with open_zip("{dest}.zip".format(dest=dest)) as zf:
        assert [] == zf.namelist()


def assert_flat(
    tmpdir,  # type: Any
    layout,  # type: InstallLayout.Value
    *extra_args  # type: str
):
    # type: (...) -> None

    dest = os.path.join(str(tmpdir), "dest")
    run_pex3(
        "venv", "create", "--layout", str(layout), "cowsay==5.0", "-d", dest, *extra_args
    ).assert_success()

    sys_path_entry = dest if layout is InstallLayout.FLAT else "{dest}.zip".format(dest=dest)
    env = make_env(PYTHONPATH=sys_path_entry)

    assert sys_path_entry == commonpath(
        [
            sys_path_entry,
            subprocess.check_output(
                args=[sys.executable, "-c", "import cowsay; print(cowsay.__file__)"], env=env
            )
            .decode("utf-8")
            .strip(),
        ]
    )

    assert (
        "5.0"
        == subprocess.check_output(args=[sys.executable, "-m", "cowsay", "--version"], env=env)
        .decode("utf-8")
        .strip()
    )

    if layout is InstallLayout.FLAT_ZIPPED:
        search_path_entry = os.path.join(str(tmpdir), "zip_contents")
        with open_zip(sys_path_entry) as zf:
            zf.extractall(search_path_entry)
    else:
        search_path_entry = dest

    assert [(ProjectName("cowsay"), Version("5.0"))] == [
        (dist.metadata.project_name, dist.metadata.version)
        for dist in dist_metadata.find_distributions(search_path=[search_path_entry])
    ]


@pytest.mark.parametrize(
    "layout",
    [
        pytest.param(layout, id=str(layout))
        for layout in (InstallLayout.FLAT, InstallLayout.FLAT_ZIPPED)
    ],
)
def test_flat(
    tmpdir,  # type: Any
    layout,  # type: InstallLayout.Value
    lock,  # type: str
    cowsay_pex,  # type: str
):
    # type: (...) -> None

    assert_flat(tmpdir, layout)
    assert_flat(tmpdir, layout, "--lock", lock)
    assert_flat(tmpdir, layout, "--pex-repository", cowsay_pex)


def test_flat_zipped_prefix(
    tmpdir,  # type: Any
    lock,  # type: str
):
    # type: (...) -> None

    dest = os.path.join(str(tmpdir), "dest")
    run_pex3(
        "venv",
        "create",
        "ansicolors",
        "--lock",
        lock,
        "--layout",
        "flat-zipped",
        "--prefix",
        "python",
        "-d",
        dest,
    ).assert_success()

    sys_path_entry = os.path.join("{dest}.zip".format(dest=dest), "python")
    assert (
        colors.cyan("ide")
        == subprocess.check_output(
            args=[sys.executable, "-c", "import colors; print(colors.cyan('ide'))"],
            env=make_env(PYTHONPATH=sys_path_entry),
        )
        .decode("utf-8")
        .strip()
    )


def test_venv_pip(tmpdir):
    # type: (Any) -> None

    dest = os.path.join(str(tmpdir), "dest")
    run_pex3("venv", "create", "-d", dest).assert_success()

    venv = Virtualenv(dest)
    assert "pip" not in [os.path.basename(exe) for exe in venv.iter_executables()]
    assert [] == list(venv.iter_distributions())

    run_pex3("venv", "create", "-d", dest, "--pip").assert_success()
    assert "pip" in [os.path.basename(exe) for exe in venv.iter_executables()]
    distributions = {
        dist.metadata.project_name: dist.metadata.version for dist in venv.iter_distributions()
    }
    pip_version = distributions[ProjectName("pip")]
    expected_prefix = "pip {version} from {prefix}".format(version=pip_version.raw, prefix=dest)
    assert (
        subprocess.check_output(args=[venv.bin_path("pip"), "--version"])
        .decode("utf-8")
        .startswith(expected_prefix)
    )


@pytest.fixture(scope="module")
def colors_pex(
    td,  # type: Any
    lock,  # type: str
):
    # type: (...) -> str

    src = str(td.join("src"))
    with safe_open(os.path.join(src, "exe.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import colors


                print(colors.magenta("Red Dwarf"))
                """
            )
        )
    pex = str(td.join("pex"))
    run_pex_command(
        args=["--lock", lock, "ansicolors", "-D", src, "-m", "exe", "-o", pex]
    ).assert_success()
    return pex


def assert_deps_only(
    interpreter,  # type: PythonInterpreter
    expected_prefix,  # type: str
    **extra_env  # type: str
):
    # type: (...) -> None

    assert expected_prefix == commonpath(
        [
            expected_prefix,
            subprocess.check_output(
                args=[
                    interpreter.binary,
                    "-c",
                    dedent(
                        """\
                        import os
                        import sys

                        import colors

                        try:
                            import exe
                            sys.exit("The deps scope should not have included the exe.py src.")
                        except ImportError:
                            pass

                        print(os.path.realpath(colors.__file__))
                        """
                    ),
                ],
                env=make_env(**extra_env),
            )
            .decode("utf-8")
            .strip(),
        ]
    )


def assert_srcs_only(
    sys_path_entry,  # type: str
    *extra_srcs  # type: str
):
    # type: (...) -> None

    assert sorted(list(extra_srcs) + ["exe.py"]) == sorted(
        os.path.relpath(os.path.join(root, f), sys_path_entry)
        for root, _, files in os.walk(sys_path_entry)
        for f in files
    )


def test_pex_scope_venv(
    tmpdir,  # type: Any
    colors_pex,  # type: str
):
    # type: (...) -> None

    dest = os.path.join(str(tmpdir), "dest")
    run_pex3(
        "venv", "create", "-d", dest, "--pex-repository", colors_pex, "--scope", "deps"
    ).assert_success()

    venv_pex_script = os.path.join(dest, "pex")
    assert not os.path.exists(venv_pex_script)

    venv = Virtualenv(dest)
    assert_deps_only(interpreter=venv.interpreter, expected_prefix=venv.site_packages_dir)

    def install_srcs():
        # type: () -> None
        run_pex3(
            "venv", "create", "-d", dest, "--pex-repository", colors_pex, "--scope", "srcs"
        ).assert_success()

    install_srcs()
    assert (
        colors.magenta("Red Dwarf")
        == subprocess.check_output(args=[venv_pex_script]).decode("utf-8").strip()
    )

    shutil.rmtree(dest)
    install_srcs()
    assert_srcs_only(
        venv.site_packages_dir,
        # Any venv created from a PEX supports the PEX_EXTRA_SYS_PATH runtime env var via this
        # `.pth` file.
        "PEX_EXTRA_SYS_PATH.pth",
    )


def test_pex_scope_flat(
    tmpdir,  # type: Any
    colors_pex,  # type: str
):
    # type: (...) -> None

    dest = os.path.join(str(tmpdir), "dest")
    run_pex3(
        "venv",
        "create",
        "-d",
        dest,
        "--pex-repository",
        colors_pex,
        "--scope",
        "deps",
        "--layout",
        "flat",
    ).assert_success()

    venv_pex_script = os.path.join(dest, "pex")
    assert not os.path.exists(venv_pex_script)
    assert_deps_only(interpreter=PythonInterpreter.get(), expected_prefix=dest, PYTHONPATH=dest)

    def install_srcs():
        # type: () -> None
        run_pex3(
            "venv",
            "create",
            "-d",
            dest,
            "--pex-repository",
            colors_pex,
            "--scope",
            "srcs",
            "--layout",
            "flat",
        ).assert_success()

    install_srcs()
    assert not os.path.exists(venv_pex_script)
    assert (
        colors.magenta("Red Dwarf")
        == subprocess.check_output(
            args=[sys.executable, "-m", "exe"], env=make_env(PYTHONPATH=dest)
        )
        .decode("utf-8")
        .strip()
    )

    shutil.rmtree(dest)
    install_srcs()
    assert_srcs_only(sys_path_entry=dest)


@pytest.fixture
def foreign_platform():
    # type: () -> str
    return "linux_x86_64-cp-310-cp310" if IS_MAC else "macosx_10.9_x86_64-cp-310-cp310"


def test_foreign_target(
    tmpdir,  # type: Any
    foreign_platform,  # type: str
):
    # type: (...) -> None

    dest = os.path.join(str(tmpdir), "dest")
    result = run_pex3(
        "venv",
        "create",
        "psutil==5.9.5",
        "-d",
        dest,
        "--platform",
        foreign_platform,
    )
    result.assert_failure()
    assert (
        "Cannot create a local venv for foreign platform {platform}.".format(
            platform=Platform.create(foreign_platform)
        )
        == result.error.strip()
    )

    run_pex3(
        "venv",
        "create",
        "--layout",
        "flat",
        "psutil==5.9.5",
        "-d",
        dest,
        "--platform",
        foreign_platform,
    ).assert_success()

    distributions = list(dist_metadata.find_distributions(search_path=[dest]))
    assert 1 == len(distributions)

    dist = distributions[0]
    assert ProjectName("psutil") == dist.metadata.project_name
    assert Version("5.9.5") == dist.metadata.version


def test_venv_update_target_mismatch(
    tmpdir,  # type: Any
    foreign_platform,  # type: str
):
    # type: (...) -> None

    dest = os.path.join(str(tmpdir), "dest")
    run_pex3("venv", "create", "-d", dest).assert_success()

    result = run_pex3(
        "venv", "create", "ansicolors==1.1.8", "-d", dest, "--platform", foreign_platform
    )
    result.assert_failure()
    assert (
        "Cannot update a local venv using a foreign platform. Given: {foreign_platform}.".format(
            foreign_platform=Platform.create(foreign_platform)
        )
        == result.error.strip()
    ), result.error

    python = ensure_python_interpreter(PY310 if sys.version_info[:2] != (3, 10) else PY39)
    result = run_pex3("venv", "create", "ansicolors==1.1.8", "-d", dest, "--python", python)
    result.assert_failure()
    assert (
        "Cannot update venv at {dest} created with {created_with} using {using}".format(
            dest=dest,
            created_with=PythonInterpreter.get().resolve_base_interpreter().binary,
            using=PythonInterpreter.from_binary(python).resolve_base_interpreter().binary,
        )
        in result.error.strip()
    ), result.error

    assert [] == list(Virtualenv(dest).iter_distributions())
    run_pex3("venv", "create", "ansicolors==1.1.8", "-d", dest).assert_success()
    assert [(ProjectName("ansicolors"), Version("1.1.8"))] == [
        (dist.metadata.project_name, dist.metadata.version)
        for dist in Virtualenv(dest).iter_distributions()
    ]



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import re
import subprocess
import sys
from textwrap import dedent

import pytest

from pex.common import safe_open
from pex.compatibility import commonpath
from pex.interpreter import PythonInterpreter
from pex.interpreter_constraints import InterpreterConstraint
from pex.pex import PEX
from pex.pex_bootstrapper import ensure_venv
from pex.pex_info import PexInfo
from pex.testing import PY38, PY39, PY_VER, ensure_python_interpreter, make_env, run_pex_command
from pex.typing import TYPE_CHECKING
from pex.venv.installer import CollisionError
from pex.venv.virtualenv import Virtualenv

if TYPE_CHECKING:
    from typing import Any, Optional, Set, Text


def test_ensure_venv_short_link(
    pex_bdist,  # type: str
    tmpdir,  # type: Any
):
    # type: (...) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")

    collision_src = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(collision_src, "will_not_collide_module.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                def verb():
                  return 42
                """
            )
        )
    with safe_open(os.path.join(collision_src, "setup.cfg"), "w") as fp:
        fp.write(
            dedent(
                """\
                [metadata]
                name = collision
                version = 0.0.1

                [options]
                py_modules =
                    will_not_collide_module
                
                [options.entry_points]
                console_scripts =
                    pex = will_not_collide_module:verb
                """
            )
        )
    with safe_open(os.path.join(collision_src, "setup.py"), "w") as fp:
        fp.write("from setuptools import setup; setup()")

    collisions_pex = os.path.join(str(tmpdir), "collisions.pex")
    run_pex_command(
        args=[
            pex_bdist,
            collision_src,
            "-o",
            collisions_pex,
            "--runtime-pex-root",
            pex_root,
            "--venv",
        ]
    ).assert_success()

    with pytest.raises(CollisionError):
        ensure_venv(PEX(collisions_pex), collisions_ok=False)

    # The directory structure for successfully executed --venv PEXes is:
    #
    # PEX_ROOT/
    #   venvs/
    #     s/  # shortcuts dir
    #       <short hash>/
    #         venv -> <real venv parent dir (see below)>
    #     <full hash1>/
    #       <full hash2>/
    #         <real venv>
    #
    # AtomicDirectory locks are used to create both branches of the venvs/ tree; so if there is a
    # failure creating a venv we expect just:
    #
    # PEX_ROOT/
    #   venvs/
    #     s/
    #       .<short hash>.atomic_directory.lck
    #     <full hash1>/
    #       .<full hash2>.atomic_directory.lck

    expected_venv_dir = PexInfo.from_pex(collisions_pex).runtime_venv_dir(collisions_pex)
    assert expected_venv_dir is not None

    full_hash1_dir = os.path.basename(os.path.dirname(expected_venv_dir))
    full_hash2_dir = os.path.basename(expected_venv_dir)

    venvs_dir = os.path.join(pex_root, "venvs")
    assert {"s", full_hash1_dir} == set(os.listdir(venvs_dir))
    short_listing = os.listdir(os.path.join(venvs_dir, "s"))
    assert 1 == len(short_listing)
    assert re.match(r"^\.[0-9a-f]+\.atomic_directory.lck", short_listing[0])
    assert [".{full_hash2}.atomic_directory.lck".format(full_hash2=full_hash2_dir)] == os.listdir(
        os.path.join(venvs_dir, full_hash1_dir)
    )

    venv_pex = ensure_venv(PEX(collisions_pex), collisions_ok=True)
    # We happen to know built distributions are always ordered before downloaded wheels in PEXes
    # as a detail of how `pex/resolver.py` works.
    assert 42 == subprocess.Popen(args=[venv_pex.pex], env=make_env(PEX_SCRIPT="pex")).wait()


def test_ensure_venv_namespace_packages(tmpdir):
    # type: (Any) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")

    # We know the twitter.common.metrics distributions depends on 4 other distributions contributing
    # to the twitter.common namespace package:
    # + twitter.common.exceptions
    # + twitter.common.decorators
    # + twitter.common.lang
    # + twitter.common.quantity
    def create_ns_pkg_pex(copies):
        # type: (bool) -> Virtualenv
        nspkgs_pex = os.path.join(
            str(tmpdir), "ns-pkgs-{style}.pex".format(style="copies" if copies else "symlinks")
        )
        run_pex_command(
            args=[
                "twitter.common.metrics==0.3.11",
                "-o",
                nspkgs_pex,
                "--runtime-pex-root",
                pex_root,
                "--venv",
                "--venv-site-packages-copies" if copies else "--no-venv-site-packages-copies",
            ]
        ).assert_success()
        nspkgs_venv_pex = ensure_venv(PEX(nspkgs_pex), collisions_ok=False)

        pex_info = PexInfo.from_pex(nspkgs_pex)
        venv_dir = pex_info.runtime_venv_dir(nspkgs_pex)
        assert venv_dir is not None
        venv = Virtualenv(venv_dir=venv_dir)
        assert os.path.realpath(nspkgs_venv_pex.pex) == os.path.realpath(venv.join_path("pex"))
        return venv

    venv_copies = create_ns_pkg_pex(copies=True)
    assert not os.path.exists(os.path.join(venv_copies.site_packages_dir, "pex-ns-pkgs.pth"))

    venv_symlinks = create_ns_pkg_pex(copies=False)
    pex_ns_pkgs_pth = os.path.join(venv_symlinks.site_packages_dir, "pex-ns-pkgs.pth")
    assert os.path.isfile(pex_ns_pkgs_pth)
    with open(pex_ns_pkgs_pth) as fp:
        assert 4 == len(fp.readlines())

    expected_path_entries = [
        os.path.join(venv_symlinks.site_packages_dir, d)
        for d in ("", "pex-ns-pkgs/1", "pex-ns-pkgs/2", "pex-ns-pkgs/3", "pex-ns-pkgs/4")
    ]
    for d in expected_path_entries:
        assert os.path.islink(os.path.join(venv_symlinks.site_packages_dir, d, "twitter"))
        assert os.path.isdir(os.path.join(venv_symlinks.site_packages_dir, d, "twitter", "common"))

    def find_package_paths(venv):
        # type: (Virtualenv) -> Set[Text]
        return set(
            subprocess.check_output(
                args=[
                    venv.join_path("pex"),
                    "-c",
                    dedent(
                        """\
                        from __future__ import print_function
                        import os
    
                        from twitter.common import decorators, exceptions, lang, metrics, quantity
        
                        
                        for pkg in decorators, exceptions, lang, metrics, quantity:
                            # These are all packages; so __file__ looks like:
                            #   <sys.path entry>/twitter/common/<pkg>/__init__.pyc
                            print(os.path.realpath(os.path.dirname(os.path.dirname(pkg.__file__))))
                        """
                    ),
                ]
            )
            .decode("utf-8")
            .splitlines()
        )

    assert 1 == len(
        find_package_paths(venv_copies)
    ), "Expected 1 unique package path for a venv built from copies."

    symlink_package_paths = find_package_paths(venv_symlinks)
    assert 5 == len(symlink_package_paths), "Expected 5 unique package paths for symlinked venv."

    # We expect package paths like:
    #   .../twitter.common.foo-0.3.11.*.whl/twitter/common
    package_file_installed_wheel_dirs = {
        os.path.dirname(os.path.dirname(p)) for p in symlink_package_paths
    }
    assert os.path.realpath(os.path.join(pex_root, PexInfo.INSTALL_CACHE)) == os.path.realpath(
        commonpath(list(package_file_installed_wheel_dirs))
    ), "Expected contributing wheel content to be symlinked from the installed wheel cache."
    assert {
        "twitter.common.{package}-0.3.11-py{py_major}-none-any.whl".format(
            package=p, py_major=venv_symlinks.interpreter.version[0]
        )
        for p in ("decorators", "exceptions", "lang", "metrics", "quantity")
    } == {
        os.path.basename(d) for d in package_file_installed_wheel_dirs
    }, "Expected 5 unique contributing wheels."


def test_ensure_venv_site_packages_copies(
    pex_bdist,  # type: str
    tmpdir,  # type: Any
):
    # type: (...) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")
    pex_file = os.path.join(str(tmpdir), "pex")

    def assert_venv_site_packages_copies(copies):
        # type: (bool) -> None
        run_pex_command(
            args=[
                pex_bdist,
                "-o",
                pex_file,
                "--pex-root",
                pex_root,
                "--runtime-pex-root",
                pex_root,
                "--venv",
                "--venv-site-packages-copies" if copies else "--no-venv-site-packages-copies",
                "--seed",
            ]
        ).assert_success()

        venv_dir = PexInfo.from_pex(pex_file).runtime_venv_dir(pex_file)
        assert venv_dir is not None
        venv = Virtualenv(venv_dir=venv_dir)
        pex_package = os.path.join(venv.site_packages_dir, "pex")
        assert os.path.isdir(pex_package)
        assert copies != os.path.islink(pex_package)

    assert_venv_site_packages_copies(copies=True)
    assert_venv_site_packages_copies(copies=False)


def test_boot_compatible_issue_1020_no_ic(tmpdir):
    # type: (Any) -> None

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(args=["psutil==5.9.0", "-o", pex]).assert_success()

    def assert_boot(python=None):
        # type: (Optional[str]) -> None
        args = [python] if python else []
        args.extend([pex, "-c", "import psutil, sys; print(sys.executable)"])
        output = subprocess.check_output(args=args, stderr=subprocess.PIPE)

        # N.B.: We expect the current interpreter the PEX was built with to be selected since the
        # PEX contains a single platform specific distribution that only works with that
        # interpreter. If the current interpreter is in a venv though, we expect the PEX bootstrap
        # to have broken out of the venv and used its base system interpreter.
        # See:
        #   https://github.com/pantsbuild/pex/pull/1130
        #   https://github.com/pantsbuild/pex/issues/1031
        assert (
            PythonInterpreter.get().resolve_base_interpreter()
            == PythonInterpreter.from_binary(
                str(output.decode("ascii").strip())
            ).resolve_base_interpreter()
        )

    assert_boot()
    assert_boot(sys.executable)

    other_interpreter = (
        ensure_python_interpreter(PY39) if PY_VER != (3, 9) else ensure_python_interpreter(PY38)
    )
    assert_boot(other_interpreter)


def test_boot_compatible_issue_1020_ic_min_compatible_build_time_hole(tmpdir):
    # type: (Any) -> None
    other_interpreter = PythonInterpreter.from_binary(
        ensure_python_interpreter(PY39) if PY_VER != (3, 9) else ensure_python_interpreter(PY38)
    )
    current_interpreter = PythonInterpreter.get()

    min_interpreter, max_interpreter = (
        (other_interpreter, current_interpreter)
        if other_interpreter.version < current_interpreter.version
        else (current_interpreter, other_interpreter)
    )
    assert min_interpreter.version < max_interpreter.version

    # Try to build a PEX that works for min and max, but only find max locally.
    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=[
            "psutil==5.9.0",
            "-o",
            pex,
            "--python-path",
            max_interpreter.binary,
            "--interpreter-constraint",
            "{python}=={major}.{minor}.*".format(
                python=max_interpreter.identity.interpreter,
                major=min_interpreter.version[0],
                minor=min_interpreter.version[1],
            ),
            "--interpreter-constraint",
            "{python}=={major}.{minor}.*".format(
                python=max_interpreter.identity.interpreter,
                major=max_interpreter.version[0],
                minor=max_interpreter.version[1],
            ),
        ]
    ).assert_success()

    # Now try to run the PEX remotely where both min and max exist.
    output = subprocess.check_output(
        args=[min_interpreter.binary, pex, "-c", "import psutil, sys; print(sys.executable)"],
        env=make_env(
            PEX_PYTHON_PATH=os.pathsep.join((min_interpreter.binary, max_interpreter.binary))
        ),
        stderr=subprocess.PIPE,
    )

    # N.B.: We expect the max interpreter the PEX was built with to be selected since the
    # PEX contains a single platform specific distribution that only works with that
    # interpreter. If the max interpreter is in a venv though, we expect the PEX bootstrap
    # to have broken out of the venv and used its base system interpreter.
    # See:
    #   https://github.com/pantsbuild/pex/pull/1130
    #   https://github.com/pantsbuild/pex/issues/1031
    assert (
        max_interpreter.resolve_base_interpreter()
        == PythonInterpreter.from_binary(
            str(output.decode("ascii").strip())
        ).resolve_base_interpreter()
    )


def test_boot_resolve_fail(
    tmpdir,  # type: Any
    py38,  # type: PythonInterpreter
    py39,  # type: PythonInterpreter
    py310,  # type: PythonInterpreter
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(args=["--python", py38.binary, "psutil==5.9.0", "-o", pex]).assert_success()

    pex_python_path = os.pathsep.join((py39.binary, py310.binary))
    process = subprocess.Popen(
        args=[py39.binary, pex, "-c", ""],
        env=make_env(PEX_PYTHON_PATH=pex_python_path),
        stderr=subprocess.PIPE,
    )
    _, stderr = process.communicate()
    assert 0 != process.returncode
    error = stderr.decode("utf-8").strip()
    pattern = re.compile(
        r"^Failed to find compatible interpreter on path {pex_python_path}.\n"
        r"\n"
        r"Examined the following interpreters:\n"
        r"1\.\)\s+{py39_exe} {py39_req}\n"
        r"2\.\)\s+{py310_exe} {py310_req}\n"
        r"\n"
        r"No interpreter compatible with the requested constraints was found:\n"
        r"\n"
        r"  A distribution for psutil could not be resolved for {py39_exe}.\n"
        r"  Found 1 distribution for psutil that do not apply:\n"
        r"  1\.\) The wheel tags for psutil 5\.9\.0 are .+ which do not match the supported tags "
        r"of {py39_exe}:\n"
        r"  cp39-cp39-.+\n"
        r"  ... \d+ more ...\n"
        r"\n"
        r"  A distribution for psutil could not be resolved for {py310_exe}.\n"
        r"  Found 1 distribution for psutil that do not apply:\n"
        r"  1\.\) The wheel tags for psutil 5\.9\.0 are .+ which do not match the supported tags "
        r"of {py310_exe}:\n"
        r"  cp310-cp310-.+\n"
        r"  ... \d+ more ...".format(
            pex_python_path=re.escape(pex_python_path),
            py39_exe=py39.binary,
            py39_req=InterpreterConstraint.exact_version(py39),
            py310_exe=py310.binary,
            py310_req=InterpreterConstraint.exact_version(py310),
        ),
    )
    assert pattern.match(error), "Got error:\n{error}\n\nExpected pattern\n{pattern}".format(
        error=error, pattern=pattern.pattern
    )



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os

from pex.common import temporary_dir
from pex.testing import PY38, ensure_python_interpreter, run_pex_command


def test_resolve_python_requires_full_version():
    # type: () -> None
    python38 = ensure_python_interpreter(PY38)
    with temporary_dir() as tmpdir:
        constraints_file = os.path.join(tmpdir, "constraints.txt")
        with open(constraints_file, "w") as fp:
            # pandas 1.0.5 has an open-ended requirement on numpy, but is in practice
            # incompatible with numpy>=1.24.0 because it drops np.bool support.
            fp.write("numpy==1.23.5")
        result = run_pex_command(
            python=python38,
            args=[
                "pandas==1.0.5",
                "--constraints",
                constraints_file,
                "--",
                "-c",
                "import pandas; print(pandas._version.get_versions()['version'])",
            ],
            quiet=True,
        )
    result.assert_success()
    assert "1.0.5" == result.output.strip()



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import subprocess
from textwrap import dedent

from pex.build_system import pep_518
from pex.build_system.pep_518 import BuildSystem
from pex.common import touch
from pex.environment import PEXEnvironment
from pex.pep_503 import ProjectName
from pex.pip.version import PipVersion
from pex.resolve.configured_resolver import ConfiguredResolver
from pex.resolve.resolver_configuration import PipConfiguration
from pex.result import Error
from pex.targets import LocalInterpreter
from pex.typing import TYPE_CHECKING
from pex.variables import ENV

if TYPE_CHECKING:
    from typing import Any, Optional, Union


def load_build_system(project_directory):
    # type: (...) -> Union[Optional[BuildSystem], Error]
    return pep_518.load_build_system(
        LocalInterpreter.create(),
        ConfiguredResolver(PipConfiguration(version=PipVersion.VENDORED)),
        project_directory,
    )


def test_load_build_system_not_a_python_project(tmpdir):
    # type: (Any) -> None
    assert load_build_system(str(tmpdir)) is None


def test_load_build_system_setup_py(tmpdir):
    # type: (Any) -> None
    project_dir = str(tmpdir)
    touch(os.path.join(project_dir, "setup.py"))
    assert load_build_system(project_dir) is None


def test_load_build_system_pyproject_but_not_for_build(tmpdir):
    # type: (Any) -> None
    project_dir = str(tmpdir)
    pyproject_toml = os.path.join(project_dir, "pyproject.toml")
    touch(pyproject_toml)
    assert load_build_system(project_dir) is None

    with open(pyproject_toml, "w") as fp:
        fp.write(
            dedent(
                """\
                [tool.black]
                target_version = ["py35"]
                """
            )
        )
    assert load_build_system(project_dir) is None


def test_load_build_system_pyproject(
    tmpdir,  # type: Any
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    build_system = load_build_system(pex_project_dir)
    assert isinstance(build_system, BuildSystem)
    assert "flit_core.buildapi" == build_system.build_backend
    dists = {
        dist.metadata.project_name
        for dist in PEXEnvironment.mount(build_system.venv_pex.pex).resolve()
    }
    assert ProjectName("flit_core") in dists
    subprocess.check_call(
        args=[build_system.venv_pex.pex, "-c", "import {}".format(build_system.build_backend)]
    )


def test_load_build_system_env_strip_issue_1872(
    tmpdir,  # type: Any
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")
    with ENV.patch(PEX_ROOT=pex_root, PEX_VERBOSE="2", PEX_SCRIPT="pex3"):
        build_system = load_build_system(pex_project_dir)
        assert isinstance(build_system, BuildSystem)
        assert pex_root == build_system.env["PEX_ROOT"]
        assert "2" == build_system.env["PEX_VERBOSE"]
        assert "PEX_SCRIPT" not in build_system.env



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os

from pex.common import temporary_dir
from pex.compatibility import to_bytes
from pex.pth import iter_pth_paths
from pex.typing import TYPE_CHECKING

try:
    from unittest import mock
except ImportError:
    import mock  # type: ignore[no-redef,import]

if TYPE_CHECKING:
    from typing import Any, Dict, List


@mock.patch("os.path.exists", autospec=True, spec_set=True)
def test_iter_pth_paths(mock_exists):
    # type: (Any) -> None
    # Ensure path checking always returns True for dummy paths.
    mock_exists.return_value = True

    with temporary_dir() as tmpdir:
        in_tmp = lambda f: os.path.join(tmpdir, f)

        PTH_TEST_MAPPING = {
            # A mapping of .pth file content -> expected paths.
            "/System/Library/Frameworks/Python.framework/Versions/2.7/Extras/lib/python\n": [
                "/System/Library/Frameworks/Python.framework/Versions/2.7/Extras/lib/python"
            ],
            "relative_path\nrelative_path2\n\nrelative_path3": [
                in_tmp("relative_path"),
                in_tmp("relative_path2"),
                in_tmp("relative_path3"),
            ],
            "duplicate_path\nduplicate_path": [in_tmp("duplicate_path")],
            "randompath\nimport nosuchmodule\n": [in_tmp("randompath")],
            "import sys\nfoo\n/bar/baz": [in_tmp("foo"), "/bar/baz"],
            "import nosuchmodule\nfoo": [],
            "import nosuchmodule\n": [],
            "import bad)syntax\n": [],
        }  # type: Dict[str, List[str]]

        for i, pth_content in enumerate(PTH_TEST_MAPPING):
            pth_tmp_path = os.path.abspath(os.path.join(tmpdir, "test%s.pth" % i))
            with open(pth_tmp_path, "wb") as f:
                f.write(to_bytes(pth_content))
            assert sorted(PTH_TEST_MAPPING[pth_content]) == sorted(
                list(iter_pth_paths(pth_tmp_path))
            )



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, division

import itertools
import os
from collections import OrderedDict, defaultdict, deque
from functools import total_ordering

from pex.common import pluralize
from pex.compatibility import url_unquote, urlparse
from pex.dist_metadata import DistMetadata, Requirement
from pex.enum import Enum
from pex.orderedset import OrderedSet
from pex.pep_425 import CompatibilityTags, TagRank
from pex.pep_503 import ProjectName
from pex.rank import Rank
from pex.requirements import VCS, VCSScheme, parse_scheme
from pex.resolve.resolved_requirement import (
    ArtifactURL,
    Fingerprint,
    PartialArtifact,
    Pin,
    ResolvedRequirement,
)
from pex.result import Error
from pex.sorted_tuple import SortedTuple
from pex.targets import LocalInterpreter, Target
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import (
        Any,
        DefaultDict,
        Deque,
        Dict,
        Iterable,
        Iterator,
        List,
        Optional,
        Protocol,
        Set,
        Tuple,
        Union,
    )

    import attr  # vendor:skip
    from packaging import tags  # vendor:skip
    from packaging import version as packaging_version  # vendor:skip
    from packaging.specifiers import SpecifierSet  # vendor:skip
else:
    from pex.third_party import attr
    from pex.third_party.packaging import tags
    from pex.third_party.packaging import version as packaging_version
    from pex.third_party.packaging.specifiers import SpecifierSet


class LockStyle(Enum["LockStyle.Value"]):
    class Value(Enum.Value):
        pass

    STRICT = Value("strict")
    SOURCES = Value("sources")
    UNIVERSAL = Value("universal")


class TargetSystem(Enum["TargetSystem.Value"]):
    class Value(Enum.Value):
        pass

    LINUX = Value("linux")
    MAC = Value("mac")
    WINDOWS = Value("windows")


@attr.s(frozen=True)
class LockConfiguration(object):
    style = attr.ib()  # type: LockStyle.Value
    requires_python = attr.ib(default=())  # type: Tuple[str, ...]
    target_systems = attr.ib(default=())  # type: Tuple[TargetSystem.Value, ...]

    @requires_python.validator
    @target_systems.validator
    def _validate_only_set_for_universal(
        self,
        attribute,  # type: Any
        value,  # type: Any
    ):
        if value and self.style != LockStyle.UNIVERSAL:
            raise ValueError(
                "The {field_name} field should only be set for {universal} style locks; "
                "this lock is {style} style and given {field_name} value of {value}".format(
                    field_name=attribute.name,
                    universal=LockStyle.UNIVERSAL.value,
                    style=self.style.value,
                    value=value,
                )
            )


@total_ordering
@attr.s(frozen=True, order=False)
class Artifact(object):
    @classmethod
    def from_artifact_url(
        cls,
        artifact_url,  # type: ArtifactURL
        fingerprint,  # type: Fingerprint
        verified=False,  # type: bool
    ):
        # type: (...) -> Union[FileArtifact, LocalProjectArtifact, VCSArtifact]
        if isinstance(artifact_url.scheme, VCSScheme):
            return VCSArtifact.from_artifact_url(
                artifact_url=artifact_url,
                fingerprint=fingerprint,
                verified=verified,
            )

        if "file" == artifact_url.scheme and os.path.isdir(artifact_url.path):
            directory = os.path.normpath(artifact_url.path)
            return LocalProjectArtifact(
                url=artifact_url.normalized_url,
                fingerprint=fingerprint,
                verified=verified,
                directory=directory,
            )

        filename = os.path.basename(artifact_url.path)
        return FileArtifact(
            url=artifact_url.normalized_url,
            fingerprint=fingerprint,
            verified=verified,
            filename=filename,
        )

    @classmethod
    def from_url(
        cls,
        url,  # type: str
        fingerprint,  # type: Fingerprint
        verified=False,  # type: bool
    ):
        # type: (...) -> Union[FileArtifact, LocalProjectArtifact, VCSArtifact]
        return cls.from_artifact_url(
            artifact_url=ArtifactURL.parse(url), fingerprint=fingerprint, verified=verified
        )

    url = attr.ib()  # type: str
    fingerprint = attr.ib()  # type: Fingerprint
    verified = attr.ib()  # type: bool

    def __lt__(self, other):
        # type: (Any) -> bool
        if not isinstance(other, Artifact):
            return NotImplemented
        return self.url < other.url


@attr.s(frozen=True, order=False)
class FileArtifact(Artifact):
    filename = attr.ib()  # type: str

    @property
    def is_source(self):
        # type: () -> bool
        return self.filename.endswith((".sdist", ".tar.gz", ".tgz", ".tar.bz2", ".tbz2", ".zip"))

    def parse_tags(self):
        # type: () -> Iterator[tags.Tag]
        if self.filename.endswith(".whl"):
            for tag in CompatibilityTags.from_wheel(self.filename):
                yield tag


@attr.s(frozen=True, order=False)
class LocalProjectArtifact(Artifact):
    directory = attr.ib()  # type: str

    @property
    def is_source(self):
        # type: () -> bool
        return True


@attr.s(frozen=True, order=False)
class VCSArtifact(Artifact):
    @classmethod
    def from_artifact_url(
        cls,
        artifact_url,  # type: ArtifactURL
        fingerprint,  # type: Fingerprint
        verified=False,  # type: bool
    ):
        # type: (...) -> VCSArtifact
        if not isinstance(artifact_url.scheme, VCSScheme):
            raise ValueError(
                "The given artifact URL is not that of a VCS artifact: {url}".format(
                    url=artifact_url.raw_url
                )
            )
        return cls(
            # N.B.: We need the raw URL in order to have access to the fragment needed for
            # `as_unparsed_requirement`.
            url=artifact_url.raw_url,
            fingerprint=fingerprint,
            verified=verified,
            vcs=artifact_url.scheme.vcs,
        )

    vcs = attr.ib()  # type: VCS.Value

    @property
    def is_source(self):
        return True

    def as_unparsed_requirement(self, project_name):
        # type: (ProjectName) -> str
        url_info = urlparse.urlparse(self.url)
        if url_info.fragment:
            fragment_parameters = urlparse.parse_qs(url_info.fragment)
            names = fragment_parameters.get("egg")
            if names and ProjectName(names[-1]) == project_name:
                # A Pip proprietary VCS requirement.
                return self.url
        # A PEP-440 direct reference VCS requirement with the project name stripped from earlier
        # processing. See: https://peps.python.org/pep-0440/#direct-references
        return "{project_name} @ {url}".format(project_name=project_name, url=self.url)


@attr.s(frozen=True)
class RankedArtifact(object):
    artifact = attr.ib()  # type: Union[FileArtifact, LocalProjectArtifact, VCSArtifact]
    rank = attr.ib()  # type: TagRank


@attr.s(frozen=True)
class LockedRequirement(object):
    @classmethod
    def create(
        cls,
        pin,  # type: Pin
        artifact,  # type: Union[FileArtifact, LocalProjectArtifact, VCSArtifact]
        requires_dists=(),  # type: Iterable[Requirement]
        requires_python=None,  # type: Optional[SpecifierSet]
        additional_artifacts=(),  # type: Iterable[Union[FileArtifact, LocalProjectArtifact, VCSArtifact]]
    ):
        # type: (...) -> LockedRequirement
        return cls(
            pin=pin,
            artifact=artifact,
            requires_dists=SortedTuple(requires_dists, key=str),
            requires_python=requires_python,
            additional_artifacts=SortedTuple(additional_artifacts),
        )

    pin = attr.ib()  # type: Pin
    artifact = attr.ib()  # type: Union[FileArtifact, LocalProjectArtifact, VCSArtifact]
    requires_dists = attr.ib(default=SortedTuple())  # type: SortedTuple[Requirement]
    requires_python = attr.ib(default=None)  # type: Optional[SpecifierSet]
    additional_artifacts = attr.ib(
        default=SortedTuple()
    )  # type: SortedTuple[Union[FileArtifact, LocalProjectArtifact, VCSArtifact]]

    def iter_artifacts(self):
        # type: () -> Iterator[Union[FileArtifact, LocalProjectArtifact, VCSArtifact]]
        yield self.artifact
        for artifact in self.additional_artifacts:
            yield artifact

    def iter_compatible_artifacts(
        self,
        target,  # type: Target
        build=True,  # type: bool
        use_wheel=True,  # type: bool
    ):
        # type: (...) -> Iterator[RankedArtifact]
        """Iterate all compatible artifacts for the given target and resolve configuration.

        :param target: The target looking to pick a resolve to use.
        :param build: Whether sdists are allowed.
        :param use_wheel: Whether wheels are allowed.
        :return: The highest ranked artifact if the requirement is compatible with the target else
            `None`.
        """
        for artifact in self.iter_artifacts():
            if build and artifact.is_source:
                # N.B.: Ensure sdists are picked last amongst a set of artifacts. We do this, since
                # a wheel is known to work with a target by the platform tags on the tin, whereas an
                # sdist may not successfully build for a given target at all. This is an affordance
                # for LockStyle.SOURCES and LockStyle.CROSS_PLATFORM lock styles.
                sdist_rank = target.supported_tags.lowest_rank.lower()
                yield RankedArtifact(artifact=artifact, rank=sdist_rank)
            elif use_wheel and isinstance(artifact, FileArtifact):
                for tag in artifact.parse_tags():
                    wheel_rank = target.supported_tags.rank(tag)
                    if wheel_rank is None:
                        continue
                    yield RankedArtifact(artifact=artifact, rank=wheel_rank)


@attr.s(frozen=True)
class _ResolveRequest(object):
    @classmethod
    def root(cls, requirement):
        # type: (Requirement) -> _ResolveRequest
        return cls(required_by=(requirement,), requirement=requirement)

    required_by = attr.ib()  # type: Tuple[Requirement, ...]
    requirement = attr.ib()  # type: Requirement
    extras = attr.ib(default=())  # type: Iterable[str]

    @property
    def project_name(self):
        # type: () -> ProjectName
        return self.requirement.project_name

    def request_dependencies(self, locked_requirement):
        # type: (LockedRequirement) -> Iterator[_ResolveRequest]
        for requires_dist in locked_requirement.requires_dists:
            yield _ResolveRequest(
                required_by=self.required_by + (requires_dist,),
                requirement=requires_dist,
                extras=self.requirement.extras,
            )

    def render_via(self):
        # type: () -> str
        return "via: {via}".format(via=" -> ".join(map(str, self.required_by)))


@attr.s(frozen=True)
class _ResolvedArtifact(object):
    ranked_artifact = attr.ib()  # type: RankedArtifact
    locked_requirement = attr.ib()  # type: LockedRequirement

    @property
    def artifact(self):
        # type: () -> Union[FileArtifact, LocalProjectArtifact, VCSArtifact]
        return self.ranked_artifact.artifact

    @property
    def version(self):
        # type: () -> Union[packaging_version.LegacyVersion, packaging_version.Version]
        return self.locked_requirement.pin.version.parsed_version

    def select_higher_rank(
        self,
        other,  # type: _ResolvedArtifact
        prefer_older_binary=False,  # type: bool
    ):
        # type: (...) -> _ResolvedArtifact

        if prefer_older_binary and self.artifact.is_source ^ other.artifact.is_source:
            return Rank.select_highest_rank(self, other, lambda ra: ra.ranked_artifact.rank)

        if self.version == other.version:
            return Rank.select_highest_rank(self, other, lambda ra: ra.ranked_artifact.rank)
        return self if self.version > other.version else other


@attr.s(frozen=True, order=False)
class _ResolvedArtifactComparator(object):
    resolved_artifact = attr.ib()  # type: _ResolvedArtifact
    prefer_older_binary = attr.ib(default=False)  # type: bool

    def __lt__(self, other):
        # type: (_ResolvedArtifactComparator) -> bool
        highest_ranked = self.resolved_artifact.select_higher_rank(
            other.resolved_artifact, self.prefer_older_binary
        )
        return highest_ranked is other.resolved_artifact


@attr.s(frozen=True)
class DownloadableArtifact(object):
    @classmethod
    def create(
        cls,
        pin,  # type: Pin
        artifact,  # type: Union[FileArtifact, LocalProjectArtifact, VCSArtifact]
        satisfied_direct_requirements=(),  # type: Iterable[Requirement]
    ):
        # type: (...) -> DownloadableArtifact
        return cls(
            pin=pin,
            artifact=artifact,
            satisfied_direct_requirements=SortedTuple(satisfied_direct_requirements, key=str),
        )

    pin = attr.ib()  # type: Pin
    artifact = attr.ib()  # type: Union[FileArtifact, LocalProjectArtifact, VCSArtifact]
    satisfied_direct_requirements = attr.ib(default=SortedTuple())  # type: SortedTuple[Requirement]


@attr.s(frozen=True)
class Resolved(object):
    @classmethod
    def create(
        cls,
        target,  # type: Target
        direct_requirements,  # type: Iterable[Requirement]
        resolved_artifacts,  # type: Iterable[_ResolvedArtifact]
        source,  # type: LockedResolve
    ):
        # type: (...) -> Resolved

        direct_requirements_by_project_name = defaultdict(
            list
        )  # type: DefaultDict[ProjectName, List[Requirement]]
        for requirement in direct_requirements:
            direct_requirements_by_project_name[requirement.project_name].append(requirement)

        # N.B.: Lowest rank means highest rank value. I.E.: The 1st tag is the most specific and
        # the 765th tag is the least specific.
        largest_rank_value = target.supported_tags.lowest_rank.value
        smallest_rank_value = TagRank.highest_natural().value
        rank_span = largest_rank_value - smallest_rank_value

        downloadable_artifacts = []
        target_specificities = []
        for resolved_artifact in resolved_artifacts:
            pin = resolved_artifact.locked_requirement.pin
            downloadable_artifacts.append(
                DownloadableArtifact.create(
                    pin=pin,
                    artifact=resolved_artifact.artifact,
                    satisfied_direct_requirements=direct_requirements_by_project_name[
                        pin.project_name
                    ],
                )
            )
            target_specificities.append(
                (rank_span - (resolved_artifact.ranked_artifact.rank.value - smallest_rank_value))
                / rank_span
            )

        return cls(
            target_specificity=sum(target_specificities) / len(target_specificities),
            downloadable_artifacts=tuple(downloadable_artifacts),
            source=source,
        )

    @classmethod
    def most_specific(cls, resolves):
        # type: (Iterable[Resolved]) -> Resolved
        sorted_resolves = sorted(resolves)
        if len(sorted_resolves) == 0:
            raise ValueError("Given no resolves to pick from.")
        # The most specific has the highest specificity which sorts last.
        return sorted_resolves[-1]

    target_specificity = attr.ib()  # type: float
    downloadable_artifacts = attr.ib()  # type: Tuple[DownloadableArtifact, ...]
    source = attr.ib(eq=False)  # type: LockedResolve


if TYPE_CHECKING:

    class Fingerprinter(Protocol):
        def fingerprint(self, artifacts):
            # type: (Iterable[PartialArtifact]) -> Iterator[FileArtifact]
            pass


@attr.s(frozen=True)
class LockedResolve(object):
    @classmethod
    def create(
        cls,
        resolved_requirements,  # type: Iterable[ResolvedRequirement]
        dist_metadatas,  # type: Iterable[DistMetadata]
        fingerprinter,  # type: Fingerprinter
        platform_tag=None,  # type: Optional[tags.Tag]
    ):
        # type: (...) -> LockedResolve

        artifacts_to_fingerprint = OrderedSet(
            itertools.chain.from_iterable(
                resolved_requirement.iter_artifacts_to_fingerprint()
                for resolved_requirement in resolved_requirements
            )
        )
        file_artifact_by_partial_artifact = dict(
            zip(
                artifacts_to_fingerprint,
                tuple(fingerprinter.fingerprint(artifacts_to_fingerprint)),
            )
        )

        def resolve_fingerprint(partial_artifact):
            # type: (PartialArtifact) -> Union[FileArtifact, LocalProjectArtifact, VCSArtifact]
            file_artifact = file_artifact_by_partial_artifact.get(partial_artifact)
            if file_artifact:
                return file_artifact
            assert partial_artifact.fingerprint is not None, (
                "No FileArtifact for {partial_artifact} mapped:\n"
                "{mapping}\n"
                "to_map:\n"
                "{to_map}".format(
                    partial_artifact=partial_artifact,
                    mapping="\n".join(map(str, file_artifact_by_partial_artifact.items())),
                    to_map="\n".join(map(str, artifacts_to_fingerprint)),
                )
            )
            return Artifact.from_artifact_url(
                artifact_url=partial_artifact.url,
                fingerprint=partial_artifact.fingerprint,
                verified=partial_artifact.verified,
            )

        dist_metadata_by_pin = {
            Pin(dist_info.project_name, dist_info.version): dist_info
            for dist_info in dist_metadatas
        }
        locked_requirements = []
        for resolved_requirement in resolved_requirements:
            distribution_metadata = dist_metadata_by_pin.get(resolved_requirement.pin)
            if distribution_metadata is None:
                raise ValueError(
                    "No distribution metadata found for {project}.\n"
                    "Given distribution metadata for:\n"
                    "{projects}".format(
                        project=resolved_requirement.pin.as_requirement(),
                        projects="\n".join(
                            sorted(str(pin.as_requirement()) for pin in dist_metadata_by_pin)
                        ),
                    )
                )
            locked_requirements.append(
                LockedRequirement.create(
                    pin=resolved_requirement.pin,
                    artifact=resolve_fingerprint(resolved_requirement.artifact),
                    requires_dists=distribution_metadata.requires_dists,
                    requires_python=distribution_metadata.requires_python,
                    additional_artifacts=(
                        resolve_fingerprint(artifact)
                        for artifact in resolved_requirement.additional_artifacts
                    ),
                )
            )
        return cls(locked_requirements=SortedTuple(locked_requirements), platform_tag=platform_tag)

    locked_requirements = attr.ib()  # type: SortedTuple[LockedRequirement]
    platform_tag = attr.ib(order=str, default=None)  # type: Optional[tags.Tag]

    def resolve(
        self,
        target,  # type: Target
        requirements,  # type: Iterable[Requirement]
        constraints=(),  # type: Iterable[Requirement]
        source=None,  # type: Optional[str]
        transitive=True,  # type: bool
        build=True,  # type: bool
        use_wheel=True,  # type: bool
        prefer_older_binary=False,  # type: bool
        include_all_matches=False,  # type: bool
    ):
        # type: (...) -> Union[Resolved, Error]

        is_local_interpreter = isinstance(target, LocalInterpreter)
        if not use_wheel and not build:
            return Error(
                "Cannot both ignore wheels (use_wheel=False) and refrain from building "
                "distributions (build=False)."
            )

        repository = defaultdict(list)  # type: DefaultDict[ProjectName, List[LockedRequirement]]
        for locked_requirement in self.locked_requirements:
            repository[locked_requirement.pin.project_name].append(locked_requirement)

        # 1. Gather all required projects and their requirers.
        required = OrderedDict()  # type: OrderedDict[ProjectName, List[_ResolveRequest]]
        to_be_resolved = deque()  # type: Deque[_ResolveRequest]

        def request_resolve(requests):
            # type: (Iterable[_ResolveRequest]) -> None
            to_be_resolved.extend(
                request
                for request in requests
                if target.requirement_applies(request.requirement, extras=request.extras)
            )

        resolved = {}  # type: Dict[ProjectName, Set[str]]
        request_resolve(_ResolveRequest.root(requirement) for requirement in requirements)
        while to_be_resolved:
            resolve_request = to_be_resolved.popleft()
            project_name = resolve_request.project_name
            required.setdefault(project_name, []).append(resolve_request)

            if not transitive:
                continue

            required_extras = set(resolve_request.requirement.extras)
            if project_name not in resolved:
                resolved[project_name] = required_extras
            else:
                resolved_extras = resolved[project_name]
                if required_extras.issubset(resolved_extras):
                    continue
                resolved_extras.update(required_extras)

            for locked_requirement in repository[project_name]:
                request_resolve(resolve_request.request_dependencies(locked_requirement))

        # 2. Select either the best fit artifact for each requirement or collect an error.
        constraints_by_project_name = {
            constraint.project_name: constraint for constraint in constraints
        }
        resolved_artifacts = []  # type: List[_ResolvedArtifact]
        errors = []
        for project_name, resolve_requests in required.items():
            reasons = []  # type: List[str]
            compatible_artifacts = []  # type: List[_ResolvedArtifact]
            for locked_requirement in repository[project_name]:

                def attributed_reason(reason):
                    # type: (str) -> str
                    if len(resolve_requests) == 1:
                        return "{pin} ({via}) {reason}".format(
                            pin=locked_requirement.pin,
                            via=resolve_requests[0].render_via(),
                            reason=reason,
                        )
                    return (
                        "{pin} {reason}\n"
                        "    requirers:\n"
                        "    {vias}".format(
                            pin=locked_requirement.pin,
                            reason=reason,
                            vias="\n    ".join(rr.render_via() for rr in resolve_requests),
                        )
                    )

                if locked_requirement.requires_python and not target.requires_python_applies(
                    locked_requirement.requires_python,
                    source=locked_requirement.pin.as_requirement(),
                ):
                    reasons.append(
                        attributed_reason(
                            "requires Python {specifier}".format(
                                specifier=locked_requirement.requires_python,
                            )
                        )
                    )
                    continue

                version_mismatches = []
                for resolve_request in resolve_requests:
                    # Pex / Pip already considered `--pre` / `--no-pre` and the rules laid out in
                    # https://peps.python.org/pep-0440/#handling-of-pre-releases during the lock
                    # resolve; so we trust that resolve's conclusion about prereleases and are
                    # permissive here.
                    if not resolve_request.requirement.contains(
                        locked_requirement.pin.version, prereleases=True
                    ):
                        version_mismatches.append(
                            "{specifier} ({via})".format(
                                specifier=resolve_request.requirement.specifier,
                                via=resolve_request.render_via(),
                            )
                        )
                constraint = constraints_by_project_name.get(locked_requirement.pin.project_name)
                if (
                    constraint is not None
                    and str(locked_requirement.pin.version) not in constraint.specifier
                ):
                    version_mismatches.append(
                        "{specifier} (via: constraint)".format(specifier=constraint.specifier)
                    )
                if version_mismatches:
                    reasons.append(
                        "{pin} does not satisfy the following requirements:\n{mismatches}".format(
                            pin=locked_requirement.pin,
                            mismatches="\n".join(
                                "    {version_mismatch}".format(version_mismatch=version_mismatch)
                                for version_mismatch in version_mismatches
                            ),
                        )
                    )
                    continue

                ranked_artifacts = tuple(
                    locked_requirement.iter_compatible_artifacts(
                        target,
                        build=build,
                        use_wheel=use_wheel,
                    )
                )
                if not ranked_artifacts:
                    reasons.append(
                        attributed_reason(
                            "does not have any compatible artifacts:\n{artifacts}".format(
                                artifacts="\n".join(
                                    "    {url}".format(url=artifact.url)
                                    for artifact in locked_requirement.iter_artifacts()
                                )
                            )
                        )
                    )
                    continue
                compatible_artifacts.extend(
                    _ResolvedArtifact(ranked_artifact, locked_requirement)
                    for ranked_artifact in ranked_artifacts
                )

            if not compatible_artifacts:
                if reasons:
                    errors.append(
                        "Dependency on {project_name} not satisfied, {count} incompatible "
                        "{candidates} found:\n{reasons}".format(
                            project_name=project_name,
                            count=len(reasons),
                            candidates=pluralize(reasons, "candidate"),
                            reasons="\n".join(
                                "{index}.) {reason}".format(index=index, reason=reason)
                                for index, reason in enumerate(reasons, start=1)
                            ),
                        )
                    )
                elif len(resolve_requests) == 1:
                    errors.append(
                        "Dependency on {project_name} ({via}) not satisfied, no candidates "
                        "found.".format(
                            project_name=project_name, via=resolve_requests[0].render_via()
                        )
                    )
                else:
                    errors.append(
                        "Dependency on {project_name} not satisfied, no candidates found:\n"
                        "    requirers:\n"
                        "    {vias}".format(
                            project_name=project_name,
                            vias="\n    ".join(rr.render_via() for rr in resolve_requests),
                        )
                    )
                continue

            compatible_artifacts.sort(
                key=lambda ra: _ResolvedArtifactComparator(
                    ra, prefer_older_binary=prefer_older_binary
                ),
                reverse=True,  # We want the highest rank sorted 1st.
            )
            if include_all_matches:
                resolved_artifacts.extend(compatible_artifacts)
            else:
                resolved_artifacts.append(compatible_artifacts[0])

        if errors:
            from_source = " from {source}".format(source=source) if source else ""
            return Error(
                "Failed to resolve all requirements for {target}{from_source}:\n"
                "\n"
                "Configured with:\n"
                "    build: {build}\n"
                "    use_wheel: {use_wheel}\n"
                "\n"
                "{errors}".format(
                    target=target.render_description(),
                    from_source=from_source,
                    build=build,
                    use_wheel=use_wheel,
                    errors="\n\n".join("{error}".format(error=error) for error in errors),
                )
            )

        uniqued_resolved_artifacts = []  # type: List[_ResolvedArtifact]
        seen = set()
        for resolved_artifact in resolved_artifacts:
            if resolved_artifact.ranked_artifact.artifact not in seen:
                uniqued_resolved_artifacts.append(resolved_artifact)
                seen.add(resolved_artifact.ranked_artifact.artifact)

        return Resolved.create(
            target=target,
            direct_requirements=requirements,
            resolved_artifacts=uniqued_resolved_artifacts,
            source=self,
        )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex import resolver
from pex.pip.version import PipVersionValue
from pex.resolve import lock_resolver
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.resolver_configuration import PipConfiguration, ReposConfiguration
from pex.resolve.resolvers import Installed, Resolver
from pex.result import try_
from pex.targets import Targets
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, Optional

    import attr  # vendor:skip
else:
    from pex.third_party import attr


_DEFAULT_REPOS = ReposConfiguration.create()


@attr.s(frozen=True)
class ConfiguredResolver(Resolver):
    @classmethod
    def default(cls):
        # type: () -> ConfiguredResolver
        return cls(PipConfiguration())

    pip_configuration = attr.ib()  # type: PipConfiguration

    def is_default_repos(self):
        return self.pip_configuration.repos_configuration == _DEFAULT_REPOS

    def resolve_lock(
        self,
        lock,  # type: Lockfile
        targets=Targets(),  # type: Targets
        pip_version=None,  # type: Optional[PipVersionValue]
    ):
        # type: (...) -> Installed
        # TODO(John Sirois): Use pip_version.
        return try_(
            lock_resolver.resolve_from_lock(
                targets=targets,
                lock=lock,
                resolver=self,
                indexes=self.pip_configuration.repos_configuration.indexes,
                find_links=self.pip_configuration.repos_configuration.find_links,
                resolver_version=self.pip_configuration.resolver_version,
                network_configuration=self.pip_configuration.network_configuration,
                build=self.pip_configuration.allow_builds,
                use_wheel=self.pip_configuration.allow_wheels,
                prefer_older_binary=self.pip_configuration.prefer_older_binary,
                use_pep517=self.pip_configuration.use_pep517,
                build_isolation=self.pip_configuration.build_isolation,
                compile=False,
                transitive=self.pip_configuration.transitive,
                verify_wheels=True,
                max_parallel_jobs=self.pip_configuration.max_jobs,
                pip_version=pip_version or self.pip_configuration.version,
            )
        )

    def resolve_requirements(
        self,
        requirements,  # type: Iterable[str]
        targets=Targets(),  # type: Targets
        pip_version=None,  # type: Optional[PipVersionValue]
    ):
        # type: (...) -> Installed
        # TODO(John Sirois): Use pip_version.
        return resolver.resolve(
            targets=targets,
            requirements=requirements,
            allow_prereleases=False,
            transitive=self.pip_configuration.transitive,
            indexes=self.pip_configuration.repos_configuration.indexes,
            find_links=self.pip_configuration.repos_configuration.find_links,
            resolver_version=self.pip_configuration.resolver_version,
            network_configuration=self.pip_configuration.network_configuration,
            build=self.pip_configuration.allow_builds,
            use_wheel=self.pip_configuration.allow_wheels,
            prefer_older_binary=self.pip_configuration.prefer_older_binary,
            use_pep517=self.pip_configuration.use_pep517,
            build_isolation=self.pip_configuration.build_isolation,
            compile=False,
            max_parallel_jobs=self.pip_configuration.max_jobs,
            ignore_errors=False,
            verify_wheels=True,
            pip_version=pip_version or self.pip_configuration.version,
            resolver=self,
        )



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
from hashlib import sha1
from textwrap import dedent

from pex.common import safe_mkdir, safe_open, temporary_dir, touch
from pex.pex import PEX
from pex.pex_builder import PEXBuilder
from pex.typing import cast
from pex.util import CacheHelper, DistributionHelper, named_temporary_file

try:
    from unittest import mock
except ImportError:
    import mock  # type: ignore[no-redef,import]


def test_access_zipped_assets():
    # type: (...) -> None
    pex_third_party_asset_dir = DistributionHelper.access_zipped_assets("pex", "third_party")

    resources = os.listdir(pex_third_party_asset_dir)
    assert (
        len(resources) > 0
    ), "The pex.third_party package should contain at least an __init__.py file."
    resources.remove("__init__.py")
    for path in resources:
        assert path in (
            "__init__.pyc",
            "__init__.pyo",
            "__pycache__",
        ), "Expected only __init__.py (and its compilations) in the pex.third_party package."


def test_hash():
    # type: () -> None
    empty_hash_digest = sha1().hexdigest()

    with named_temporary_file() as fp:
        fp.flush()
        assert empty_hash_digest == CacheHelper.hash(fp.name)

    with named_temporary_file() as fp:
        string = b"asdf" * 1024 * sha1().block_size + b"extra padding"
        fp.write(string)
        fp.flush()
        assert sha1(string).hexdigest() == CacheHelper.hash(fp.name)

    with named_temporary_file() as fp:
        empty_hash = sha1()
        fp.write(b"asdf")
        fp.flush()
        hash_output = CacheHelper.hash(fp.name, digest=empty_hash)
        assert hash_output == empty_hash.hexdigest()


def test_dir_hash():
    # type: () -> None
    with temporary_dir() as tmp_dir:
        safe_mkdir(os.path.join(tmp_dir, "a", "b"))
        with safe_open(os.path.join(tmp_dir, "c", "d", "e.py"), "w") as fp:
            fp.write("contents1")
        with safe_open(os.path.join(tmp_dir, "f.py"), "w") as fp:
            fp.write("contents2")
        hash1 = CacheHelper.dir_hash(tmp_dir)

        os.rename(os.path.join(tmp_dir, "c"), os.path.join(tmp_dir, "c-renamed"))
        assert hash1 != CacheHelper.dir_hash(tmp_dir)

        os.rename(os.path.join(tmp_dir, "c-renamed"), os.path.join(tmp_dir, "c"))
        assert hash1 == CacheHelper.dir_hash(tmp_dir)

        touch(os.path.join(tmp_dir, "c", "d", "e.pyc"))
        assert hash1 == CacheHelper.dir_hash(tmp_dir)
        touch(os.path.join(tmp_dir, "c", "d", "e.pyc.123456789"))
        assert hash1 == CacheHelper.dir_hash(tmp_dir)

        pycache_dir = os.path.join(tmp_dir, "__pycache__")
        safe_mkdir(pycache_dir)
        touch(os.path.join(pycache_dir, "f.pyc"))
        assert hash1 == CacheHelper.dir_hash(tmp_dir)
        touch(os.path.join(pycache_dir, "f.pyc.123456789"))
        assert hash1 == CacheHelper.dir_hash(tmp_dir)

        touch(os.path.join(pycache_dir, "f.py"))
        assert hash1 == CacheHelper.dir_hash(
            tmp_dir
        ), "All content under __pycache__ directories should be ignored."


try:
    import __builtin__ as python_builtins  # type: ignore[import]
except ImportError:
    import builtins as python_builtins  # type: ignore[no-redef]


def assert_access_zipped_assets(distribution_helper_import):
    # type: (str) -> bytes
    test_executable = dedent(
        """
        import os
        {distribution_helper_import}
        temp_dir = DistributionHelper.access_zipped_assets('my_package', 'submodule')
        with open(os.path.join(temp_dir, 'mod.py'), 'r') as fp:
            for line in fp:
                print(line)
        """.format(
            distribution_helper_import=distribution_helper_import
        )
    )
    with temporary_dir() as td1, temporary_dir() as td2:
        pb = PEXBuilder(path=td1)
        with open(os.path.join(td1, "exe.py"), "w") as fp:
            fp.write(test_executable)
            pb.set_executable(fp.name)

        submodule = os.path.join(td1, "my_package", "submodule")
        safe_mkdir(submodule)
        mod_path = os.path.join(submodule, "mod.py")
        with open(mod_path, "w") as fp:
            fp.write("accessed")
            pb.add_source(fp.name, "my_package/submodule/mod.py")
        pb.add_source(None, "my_package/__init__.py")
        pb.add_source(None, "my_package/submodule/__init__.py")
        pex = os.path.join(td2, "app.pex")
        pb.build(pex)

        process = PEX(pex, interpreter=pb.interpreter).run(
            blocking=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        stdout, stderr = process.communicate()
        assert process.returncode == 0
        assert b"accessed\n" == stdout
        return cast(bytes, stderr)


def test_access_zipped_assets_integration():
    # type: () -> None
    stderr = assert_access_zipped_assets("from pex.util import DistributionHelper")
    assert b"" == stderr.strip()


def test_named_temporary_file():
    # type: () -> None
    with named_temporary_file() as fp:
        name = fp.name
        fp.write(b"hi")
        fp.flush()
        assert os.path.exists(name)
        with open(name) as new_fp:
            assert new_fp.read() == "hi"

    assert not os.path.exists(name)



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import sys
from argparse import Action, ArgumentError, ArgumentParser, ArgumentTypeError, _ActionsContainer
from collections import OrderedDict
from operator import attrgetter

from pex import pex_warnings
from pex.argparse import HandleBoolAction
from pex.cli.command import BuildTimeCommand
from pex.commands.command import JsonMixin, OutputMixin
from pex.common import pluralize
from pex.dist_metadata import Requirement, RequirementParseError
from pex.enum import Enum
from pex.orderedset import OrderedSet
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.resolve import requirement_options, resolver_options, target_options
from pex.resolve.config import finalize as finalize_resolve_config
from pex.resolve.locked_resolve import LockConfiguration, LockStyle, Resolved, TargetSystem
from pex.resolve.lockfile import json_codec
from pex.resolve.lockfile.create import create
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.lockfile.subset import subset
from pex.resolve.lockfile.updater import (
    ArtifactUpdate,
    FingerprintUpdate,
    LockUpdater,
    ResolveUpdateRequest,
    VersionUpdate,
)
from pex.resolve.path_mappings import PathMappings
from pex.resolve.requirement_configuration import RequirementConfiguration
from pex.resolve.resolved_requirement import Fingerprint, Pin
from pex.resolve.resolver_options import parse_lockfile
from pex.resolve.target_configuration import InterpreterConstraintsNotSatisfied, TargetConfiguration
from pex.result import Error, Ok, Result, try_
from pex.sorted_tuple import SortedTuple
from pex.targets import Targets
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING
from pex.version import __version__

if TYPE_CHECKING:
    from typing import IO, Dict, Iterable, List, Optional, Tuple, Union

    import attr  # vendor:skip
else:
    from pex.third_party import attr


class FingerprintMismatch(Enum["FingerprintMismatch.Value"]):
    class Value(Enum.Value):
        pass

    IGNORE = Value("ignore")
    WARN = Value("warn")
    ERROR = Value("error")


class ExportFormat(Enum["ExportFormat.Value"]):
    class Value(Enum.Value):
        pass

    PIP = Value("pip")
    PEP_665 = Value("pep-665")


class ExportSortBy(Enum["ExportSortBy.Value"]):
    class Value(Enum.Value):
        pass

    SPECIFICITY = Value("specificity")
    PROJECT_NAME = Value("project-name")


class DryRunStyle(Enum["DryRunStyle.Value"]):
    class Value(Enum.Value):
        pass

    DISPLAY = Value("display")
    CHECK = Value("check")


class HandleDryRunAction(Action):
    def __init__(self, *args, **kwargs):
        kwargs["nargs"] = "?"
        super(HandleDryRunAction, self).__init__(*args, **kwargs)

    def __call__(self, parser, namespace, value, option_str=None):
        if option_str.startswith("--no-"):
            if value:
                raise ArgumentError(
                    None,
                    "The {option} option does not take a value; given: {value!r}".format(
                        option=option_str, value=value
                    ),
                )
            dry_run_style = None
        elif value:
            try:
                dry_run_style = DryRunStyle.for_value(value)
            except ValueError:
                raise ArgumentTypeError(
                    "Invalid value for {option}: {value!r}. Either pass no value for {default!r} "
                    "or one of: {choices}".format(
                        option=option_str,
                        value=value,
                        default=DryRunStyle.DISPLAY,
                        choices=", ".join(map(repr, DryRunStyle.values())),
                    )
                )
        else:
            dry_run_style = DryRunStyle.DISPLAY
        setattr(namespace, self.dest, dry_run_style)


class Lock(OutputMixin, JsonMixin, BuildTimeCommand):
    """Operate on PEX lock files."""

    @staticmethod
    def _add_target_options(parser):
        # type: (_ActionsContainer) -> None
        target_options.register(
            parser.add_argument_group(
                title="Target options",
                description=(
                    "Specify which interpreters and platforms resolved distributions must support."
                ),
            )
        )

    @classmethod
    def _create_resolver_options_group(cls, parser):
        # type: (_ActionsContainer) -> _ActionsContainer
        return parser.add_argument_group(
            title="Resolver options",
            description="Configure how third party distributions are resolved.",
        )

    @classmethod
    def _add_resolve_options(cls, parser):
        # type: (_ActionsContainer) -> None
        requirement_options.register(
            parser.add_argument_group(
                title="Requirement options",
                description="Indicate which third party distributions should be resolved",
            )
        )
        cls._add_target_options(parser)
        resolver_options.register(
            cls._create_resolver_options_group(parser),
            include_pex_repository=False,
            include_lock=False,
        )

    @classmethod
    def _add_lockfile_option(
        cls,
        parser,  # type: _ActionsContainer
        verb,  # type: str
        positional=True,  # type: bool
    ):
        # type: (...) -> None
        if positional:
            parser.add_argument(
                "lockfile",
                nargs=1,
                help="The Pex lock file to {verb}".format(verb=verb),
            )
        else:
            parser.add_argument(
                "--lock",
                required=True,
                help="The Pex lock file to {verb}".format(verb=verb),
            )

    @classmethod
    def _add_lock_options(cls, parser):
        # type: (_ActionsContainer) -> None
        resolver_options.register_lock_options(parser)

    @classmethod
    def _add_create_arguments(cls, create_parser):
        # type: (_ActionsContainer) -> None
        create_parser.add_argument(
            "--style",
            default=LockStyle.STRICT,
            choices=LockStyle.values(),
            type=LockStyle.for_value,
            help=(
                "The style of lock to generate. The {strict!r} style is the default and generates "
                "a lock file that contains exactly the distributions that would be used in a local "
                "PEX build. If an sdist would be used, the sdist is included, but if a wheel would "
                "be used, an accompanying sdist will not be included. The {sources!r} style "
                "includes locks containing both wheels and the associated sdists when available. "
                "The {universal!r} style generates a universal lock for all possible target "
                "interpreters and platforms, although the scope can be constrained via one or more "
                "--interpreter-constraint. Of the three lock styles, only {strict!r} can give you "
                "full confidence in the lock since it includes exactly the artifacts that are "
                "included in the local PEX you'll build to test the lock result with before "
                "checking in the lock. With the other two styles you lock un-vetted artifacts in "
                "addition to the {strict!r} ones; so, even though you can be sure to reproducibly "
                "resolve those same un-vetted artifacts in the future, they're still un-vetted and "
                "could be innocently or maliciously different from the {strict!r} artifacts you "
                "can locally vet before committing the lock to version control. The effects of the "
                "differences could range from failing a resolve using the lock when the un-vetted "
                "artifacts have different dependencies from their sibling artifacts, to your "
                "application crashing due to different code in the sibling artifacts to being "
                "compromised by differing code in the sibling artifacts. So, although the more "
                "permissive lock styles will allow the lock to work on a wider range of machines /"
                "are apparently more convenient, the convenience comes with a potential price and "
                "using these styles should be considered carefully.".format(
                    strict=LockStyle.STRICT,
                    sources=LockStyle.SOURCES,
                    universal=LockStyle.UNIVERSAL,
                )
            ),
        )
        create_parser.add_argument(
            "--target-system",
            dest="target_systems",
            default=[],
            action="append",
            choices=TargetSystem.values(),
            type=TargetSystem.for_value,
            help=(
                "The target operating systems to generate the lock for. This option applies only "
                "to `--style {universal}` locks and restricts the locked artifacts to those "
                "compatible with the specified target operating systems. By default, {universal!r} "
                "style locks include artifacts for all operating systems.".format(
                    universal=LockStyle.UNIVERSAL,
                )
            ),
        )
        cls._add_lock_options(create_parser)
        cls.add_output_option(create_parser, entity="lock")
        cls.add_json_options(create_parser, entity="lock", include_switch=False)
        cls._add_resolve_options(create_parser)

    @classmethod
    def _add_export_arguments(
        cls,
        export_parser,  # type: _ActionsContainer
        lockfile_option_positional=True,  # type: bool
    ):
        # type: (...) -> None
        export_parser.add_argument(
            "--format",
            default=ExportFormat.PIP,
            choices=ExportFormat.values(),
            type=ExportFormat.for_value,
            help=(
                "The format to export the lock to. Currently only the {pip!r} requirements file "
                "format using `--hash` is supported.".format(pip=ExportFormat.PIP)
            ),
        )
        export_parser.add_argument(
            "--sort-by",
            default=ExportSortBy.SPECIFICITY,
            choices=ExportSortBy.values(),
            type=ExportSortBy.for_value,
            help="How to sort the requirements in the export (if supported).",
        )
        cls._add_lockfile_option(
            export_parser, verb="export", positional=lockfile_option_positional
        )
        cls._add_lock_options(export_parser)
        cls.add_output_option(export_parser, entity="lock")
        cls._add_target_options(export_parser)
        resolver_options_parser = cls._create_resolver_options_group(export_parser)
        resolver_options.register_network_options(resolver_options_parser)

    @classmethod
    def _add_export_subset_arguments(cls, export_subset_parser):
        # type: (_ActionsContainer) -> None
        cls._add_export_arguments(export_subset_parser, lockfile_option_positional=False)
        requirement_options.register(export_subset_parser)

    @classmethod
    def _add_update_arguments(cls, update_parser):
        # type: (_ActionsContainer) -> None
        update_parser.add_argument(
            "-p",
            "--project",
            dest="projects",
            action="append",
            default=[],
            type=str,
            help=(
                "Just attempt to update these projects in the lock, leaving all others unchanged. "
                "If the projects aren't already in the lock, attempt to add them as top-level"
                "requirements leaving all others unchanged."
            ),
        )
        update_parser.add_argument(
            "--strict",
            "--no-strict",
            "--non-strict",
            action=HandleBoolAction,
            default=True,
            type=bool,
            help=(
                "Require all target platforms in the lock be updated at once. If any target "
                "platform in the lock file does not have a representative local interpreter to "
                "execute the lock update with, the update will fail."
            ),
        )

        update_parser.add_argument(
            "--pin",
            action=HandleBoolAction,
            default=False,
            type=bool,
            help=(
                "When performing the update, pin all projects in the lock to their current "
                "versions. This is useful to pick up newly published wheels for those projects or "
                "else switch repositories from the original ones when used in conjunction with any "
                "of --index, --no-pypi and --find-links."
            ),
        )

        update_parser.add_argument(
            "--fingerprint-mismatch",
            default=FingerprintMismatch.ERROR,
            choices=FingerprintMismatch.values(),
            type=FingerprintMismatch.for_value,
            help=(
                "What to do when a lock update would result in at least one artifact fingerprint "
                "changing: {ignore!r} the mismatch and use the new fingerprint, {warn!r} about the "
                "mismatch but use the new fingerprint anyway or {error!r} and refuse to use the "
                "new mismatching fingerprint".format(
                    ignore=FingerprintMismatch.IGNORE,
                    warn=FingerprintMismatch.WARN,
                    error=FingerprintMismatch.ERROR,
                )
            ),
        )
        update_parser.add_argument(
            "-n",
            "--dry-run",
            "--no-dry-run",
            action=HandleDryRunAction,
            help=(
                "Don't update the lock file; just report what updates would be made. By default, "
                "the report is to STDOUT and the exit code is zero. If a value of {check!r} is "
                "passed, the report is to STDERR and the exit code is non-zero.".format(
                    check=DryRunStyle.CHECK
                )
            ),
        )
        cls._add_lockfile_option(update_parser, verb="update")
        cls._add_lock_options(update_parser)
        cls.add_json_options(update_parser, entity="lock", include_switch=False)
        cls._add_target_options(update_parser)
        resolver_options_parser = cls._create_resolver_options_group(update_parser)
        resolver_options.register_repos_options(resolver_options_parser)
        resolver_options.register_network_options(resolver_options_parser)
        resolver_options.register_max_jobs_option(resolver_options_parser)

    @classmethod
    def add_extra_arguments(
        cls,
        parser,  # type: ArgumentParser
    ):
        # type: (...) -> None
        subcommands = cls.create_subcommands(
            parser,
            description="PEX lock files can be operated on using any of the following subcommands.",
        )
        with subcommands.parser(
            name="create", help="Create a lock file.", func=cls._create
        ) as create_parser:
            cls._add_create_arguments(create_parser)
        with subcommands.parser(
            name="export",
            help="Export a Pex lock file for a single targeted environment in a different format.",
            func=cls._export,
        ) as export_parser:
            cls._add_export_arguments(export_parser)
        with subcommands.parser(
            name="export-subset",
            help=(
                "Export a subset of a Pex lock file for a single targeted environment in a "
                "different format."
            ),
            func=cls._export_subset,
        ) as export_subset_parser:
            cls._add_export_subset_arguments(export_subset_parser)
        with subcommands.parser(
            name="update", help="Update a Pex lock file.", func=cls._update
        ) as update_parser:
            cls._add_update_arguments(update_parser)

    def _resolve_targets(
        self,
        action,  # type: str
        style,  # type: LockStyle.Value
        target_configuration=None,  # type: Optional[TargetConfiguration]
    ):
        # type: (...) -> Union[Targets, Error]

        target_config = target_configuration or target_options.configure(self.options)
        if style is not LockStyle.UNIVERSAL:
            return target_config.resolve_targets()

        if target_config.pythons:
            return Error(
                "When {action} a {universal} lock, the interpreters the resulting lock applies "
                "to can only be constrained via --interpreter-constraint. There {were} "
                "{num_pythons} --python specified.".format(
                    action=action,
                    universal=LockStyle.UNIVERSAL.value,
                    were="were" if len(target_config.pythons) > 1 else "was",
                    num_pythons=len(target_config.pythons),
                )
            )

        if not target_config.interpreter_constraints:
            return Targets(
                platforms=target_config.platforms,
                complete_platforms=target_config.complete_platforms,
                assume_manylinux=target_config.assume_manylinux,
            )

        try:
            interpreter = next(target_config.interpreter_configuration.iter_interpreters())
        except InterpreterConstraintsNotSatisfied as e:
            return Error(
                "When {action} a universal lock with an --interpreter-constraint, an "
                "interpreter matching the constraint must be found on the local system but "
                "none was: {err}".format(action=action, err=e)
            )
        return Targets(
            interpreters=(interpreter,),
            platforms=target_config.platforms,
            complete_platforms=target_config.complete_platforms,
            assume_manylinux=target_config.assume_manylinux,
        )

    def _create(self):
        # type: () -> Result
        target_configuration = target_options.configure(self.options)
        if self.options.style == LockStyle.UNIVERSAL:
            lock_configuration = LockConfiguration(
                style=LockStyle.UNIVERSAL,
                requires_python=tuple(
                    str(interpreter_constraint.requires_python)
                    for interpreter_constraint in target_configuration.interpreter_constraints
                ),
                target_systems=tuple(self.options.target_systems),
            )
        elif self.options.target_systems:
            return Error(
                "The --target-system option only applies to --style {universal} locks.".format(
                    universal=LockStyle.UNIVERSAL.value
                )
            )
        else:
            lock_configuration = LockConfiguration(style=self.options.style)

        requirement_configuration = requirement_options.configure(self.options)
        targets = try_(
            self._resolve_targets(
                action="creating",
                style=self.options.style,
                target_configuration=target_configuration,
            )
        )
        pip_configuration = try_(
            finalize_resolve_config(
                resolver_configuration=resolver_options.create_pip_configuration(self.options),
                targets=targets,
                context="lock creation",
            )
        )
        self._dump_lockfile(
            try_(
                create(
                    lock_configuration=lock_configuration,
                    requirement_configuration=requirement_configuration,
                    targets=targets,
                    pip_configuration=pip_configuration,
                )
            )
        )
        return Ok()

    def _load_lockfile(self):
        # type: () -> Tuple[str, Lockfile]
        lock_file_path = self.options.lock if "lock" in self.options else self.options.lockfile[0]
        return lock_file_path, try_(parse_lockfile(self.options, lock_file_path=lock_file_path))

    def _get_path_mappings(self):
        # type: () -> PathMappings
        return resolver_options.get_path_mappings(self.options)

    def _dump_lockfile(
        self,
        lock_file,  # type: Lockfile
        output=None,  # type: Optional[IO]
    ):
        # type: (...) -> None
        def dump_with_terminating_newline(out):
            # json.dump() does not write the newline terminating the last line, but some
            # JSON linters, and line-based tools in general, expect it, and since these
            # files are intended to be checked in to repos that may enforce this, we oblige.
            self.dump_json(
                self.options,
                json_codec.as_json_data(
                    lockfile=lock_file, path_mappings=self._get_path_mappings()
                ),
                out=out,
                sort_keys=True,
            )
            out.write("\n")

        if output:
            dump_with_terminating_newline(out=output)
        else:
            with self.output(self.options) as output:
                dump_with_terminating_newline(out=output)

    def _export(self, requirement_configuration=RequirementConfiguration()):
        # type: (RequirementConfiguration) -> Result
        if self.options.format != ExportFormat.PIP:
            return Error(
                "Only the {pip!r} lock format is supported currently.".format(pip=ExportFormat.PIP)
            )

        lockfile_path, lock_file = self._load_lockfile()
        targets = target_options.configure(self.options).resolve_targets()
        target = targets.require_unique_target(
            purpose="exporting a lock in the {pip!r} format".format(pip=ExportFormat.PIP)
        )

        network_configuration = resolver_options.create_network_configuration(self.options)
        with TRACER.timed("Selecting locks for {target}".format(target=target)):
            subset_result = try_(
                subset(
                    targets=targets,
                    lock=lock_file,
                    requirement_configuration=requirement_configuration,
                    network_configuration=network_configuration,
                    build=lock_file.allow_builds,
                    use_wheel=lock_file.allow_wheels,
                    prefer_older_binary=lock_file.prefer_older_binary,
                    transitive=lock_file.transitive,
                    include_all_matches=True,
                )
            )

        if len(subset_result.subsets) != 1:
            resolved = Resolved.most_specific(
                resolved_subset.resolved for resolved_subset in subset_result.subsets
            )
            pex_warnings.warn(
                "Only a single lock can be exported in the {pip!r} format.\n"
                "There were {count} locks stored in {lockfile} that were applicable for the "
                "selected target: {target}; so using the most specific lock with platform "
                "{platform}.".format(
                    count=len(subset_result.subsets),
                    lockfile=lock_file.source,
                    pip=ExportFormat.PIP,
                    target=target,
                    platform=resolved.source.platform_tag,
                )
            )
        else:
            resolved = subset_result.subsets[0].resolved

        fingerprints_by_pin = OrderedDict()  # type: OrderedDict[Pin, List[Fingerprint]]
        for downloaded_artifact in resolved.downloadable_artifacts:
            fingerprints_by_pin.setdefault(downloaded_artifact.pin, []).append(
                downloaded_artifact.artifact.fingerprint
            )

        with self.output(self.options) as output:
            pins = fingerprints_by_pin.keys()  # type: Iterable[Pin]
            if self.options.sort_by == ExportSortBy.PROJECT_NAME:
                pins = sorted(pins, key=attrgetter("project_name.normalized"))
            for pin in pins:
                fingerprints = fingerprints_by_pin[pin]
                output.write(
                    "{project_name}=={version} \\\n"
                    "  {hashes}\n".format(
                        project_name=pin.project_name,
                        version=pin.version.raw,
                        hashes=" \\\n  ".join(
                            "--hash={algorithm}:{hash}".format(
                                algorithm=fingerprint.algorithm, hash=fingerprint.hash
                            )
                            for fingerprint in fingerprints
                        ),
                    )
                )
        return Ok()

    def _export_subset(self):
        requirement_configuration = requirement_options.configure(self.options)
        return self._export(requirement_configuration=requirement_configuration)

    def _update(self):
        # type: () -> Result
        try:
            update_requirements = tuple(
                Requirement.parse(project) for project in self.options.projects
            )
        except RequirementParseError as e:
            return Error("Failed to parse project requirement to update: {err}".format(err=e))

        lock_file_path, lock_file = self._load_lockfile()
        network_configuration = resolver_options.create_network_configuration(self.options)
        lock_updater = LockUpdater.create(
            lock_file=lock_file,
            repos_configuration=resolver_options.create_repos_configuration(self.options),
            network_configuration=network_configuration,
            max_jobs=resolver_options.get_max_jobs_value(self.options),
        )

        target_configuration = target_options.configure(self.options)
        targets = try_(
            self._resolve_targets(
                action="updating", style=lock_file.style, target_configuration=target_configuration
            )
        )

        lock_updater = attr.evolve(
            lock_updater,
            pip_configuration=try_(
                finalize_resolve_config(
                    resolver_configuration=lock_updater.pip_configuration,
                    targets=targets,
                    context="lock updating",
                    pip_version=lock_file.pip_version,
                )
            ),
        )

        with TRACER.timed("Selecting locks to update"):
            subset_result = try_(
                subset(
                    targets=targets,
                    lock=lock_file,
                    network_configuration=network_configuration,
                    build=lock_file.allow_builds,
                    use_wheel=lock_file.allow_wheels,
                    prefer_older_binary=lock_file.prefer_older_binary,
                    transitive=lock_file.transitive,
                )
            )

        update_requests = [
            ResolveUpdateRequest(
                target=resolved_subset.target, locked_resolve=resolved_subset.resolved.source
            )
            for resolved_subset in subset_result.subsets
        ]
        if self.options.strict and lock_file.style is not LockStyle.UNIVERSAL:
            missing_updates = set(lock_file.locked_resolves) - {
                update_request.locked_resolve for update_request in update_requests
            }
            if missing_updates:
                return Error(
                    "This lock update is --strict but the following platforms present in "
                    "{lock_file} were not found on the local machine:\n"
                    "{missing_platforms}\n"
                    "You might be able to correct this by adjusting target options like "
                    "--python-path or else by relaxing the update to be --non-strict.".format(
                        lock_file=lock_file_path,
                        missing_platforms="\n".join(
                            sorted(
                                "+ {platform}".format(platform=locked_resolve.platform_tag)
                                for locked_resolve in missing_updates
                            )
                        ),
                    )
                )

        if not update_requests:
            return Ok(
                "No lock update was performed.\n"
                "The following platforms present in {lock_file} were not found on the local "
                "machine:\n"
                "{missing_platforms}\n"
                "You might still be able to update the lock by adjusting target options like "
                "--python-path.".format(
                    lock_file=lock_file_path,
                    missing_platforms="\n".join(
                        sorted(
                            "+ {platform}".format(platform=locked_resolve.platform_tag)
                            for locked_resolve in lock_file.locked_resolves
                        )
                    ),
                )
            )

        lock_update = try_(
            lock_updater.update(
                update_requests=update_requests,
                updates=update_requirements,
                pin=self.options.pin,
                assume_manylinux=targets.assume_manylinux,
            )
        )

        constraints_by_project_name = {
            constraint.project_name: constraint for constraint in lock_file.constraints
        }
        dry_run = self.options.dry_run
        path_mappings = self._get_path_mappings()
        output = sys.stdout if dry_run is DryRunStyle.DISPLAY else sys.stderr
        updates = []
        warnings = []  # type: List[str]
        for resolve_update in lock_update.resolves:
            platform = resolve_update.updated_resolve.platform_tag or "universal"
            print("Updates for lock generated by {platform}".format(platform=platform), file=output)
            fingerprint_updates = {}  # type: Dict[ProjectName, Version]
            for project_name, update in resolve_update.updates.items():
                if not update:
                    print(
                        "  There {tense} no updates for {project_name}".format(
                            tense="would be" if dry_run else "were",
                            project_name=project_name,
                        ),
                        file=output,
                    )
                    continue

                updates.append(update)
                if isinstance(update, VersionUpdate):
                    if update.original:
                        print(
                            "  {lead_in} {project_name} from {original_version} to "
                            "{updated_version}".format(
                                lead_in="Would update" if dry_run else "Updated",
                                project_name=project_name,
                                original_version=update.original,
                                updated_version=update.updated,
                            ),
                            file=output,
                        )
                    else:
                        print(
                            "  {lead_in} {project_name} {updated_version}".format(
                                lead_in="Would add" if dry_run else "Added",
                                project_name=project_name,
                                updated_version=update.updated,
                            ),
                            file=output,
                        )
                else:
                    message_lines = [
                        "  {lead_in} {project_name} {version} artifacts:".format(
                            lead_in="Would update" if dry_run else "Updated",
                            project_name=project_name,
                            version=update.version,
                        )
                    ]
                    if update.added:
                        message_lines.extend(
                            "    + {added}".format(
                                added=path_mappings.maybe_canonicalize(artifact.url)
                            )
                            for artifact in update.added
                        )
                    if update.updated:
                        if any(
                            isinstance(change, (FingerprintUpdate, ArtifactUpdate))
                            for change in update.updated
                        ):
                            fingerprint_updates[project_name] = update.version
                        message_lines.extend(
                            "    {changed}".format(
                                changed=path_mappings.maybe_canonicalize(change.render_update())
                            )
                            for change in update.updated
                        )
                    if update.removed:
                        message_lines.extend(
                            "    - {removed}".format(
                                removed=path_mappings.maybe_canonicalize(artifact.url)
                            )
                            for artifact in update.removed
                        )

                    print("\n".join(message_lines), file=output)
            if fingerprint_updates:
                warnings.append(
                    "Detected fingerprint changes in the following locked {projects} for lock "
                    "generated by {platform}!\n{ids}".format(
                        platform=platform,
                        projects=pluralize(fingerprint_updates, "project"),
                        ids="\n".join(
                            "{project_name} {version}".format(
                                project_name=project_name, version=version
                            )
                            for project_name, version in fingerprint_updates.items()
                        ),
                    )
                )

        if not updates:
            return Ok()

        if warnings:
            if self.options.fingerprint_mismatch in (
                FingerprintMismatch.WARN,
                FingerprintMismatch.ERROR,
            ):
                message = "\n".join(warnings)
                if self.options.fingerprint_mismatch is FingerprintMismatch.ERROR:
                    return Error(message)
                pex_warnings.warn(message)

        if dry_run:
            return Error() if dry_run is DryRunStyle.CHECK else Ok()

        original_locked_project_names = {
            locked_requirement.pin.project_name
            for locked_resolve in lock_file.locked_resolves
            for locked_requirement in locked_resolve.locked_requirements
        }
        new_requirements = OrderedSet(
            update
            for update in update_requirements
            if update.project_name not in original_locked_project_names
        )
        constraints_by_project_name.update(
            (constraint.project_name, constraint) for constraint in update_requirements
        )
        for requirement in new_requirements:
            constraints_by_project_name.pop(requirement.project_name, None)
        requirements = OrderedSet(lock_file.requirements)
        requirements.update(new_requirements)

        with open(lock_file_path, "w") as fp:
            self._dump_lockfile(
                lock_file=attr.evolve(
                    lock_file,
                    pex_version=__version__,
                    requirements=SortedTuple(requirements, key=str),
                    constraints=SortedTuple(constraints_by_project_name.values(), key=str),
                    locked_resolves=SortedTuple(
                        resolve_update.updated_resolve for resolve_update in lock_update.resolves
                    ),
                ),
                output=fp,
            )
        return Ok()



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Optional, Text, Union

    import attr  # vendor:skip
    from packaging import utils as packaging_utils  # vendor:skip
    from packaging import version as packaging_version  # vendor:skip

    ParsedVersion = Union[packaging_version.LegacyVersion, packaging_version.Version]
else:
    from pex.third_party import attr
    from pex.third_party.packaging import utils as packaging_utils
    from pex.third_party.packaging import version as packaging_version


def _ensure_ascii_str(text):
    # type: (Text) -> str

    # Version numbers must be composed of ASCII as spelled out here:
    #  https://peps.python.org/pep-0440/#summary-of-changes-to-pep-440
    return str(text)


@attr.s(frozen=True, order=True)
class Version(object):
    """A PEP-440 normalized version: https://www.python.org/dev/peps/pep-0440/#normalization"""

    raw = attr.ib(eq=False, converter=_ensure_ascii_str)  # type: str
    normalized = attr.ib(init=False)  # type: str
    _parsed_version = attr.ib(
        default=None, init=False, eq=False, repr=False
    )  # type: Optional[ParsedVersion]

    def __attrs_post_init__(self):
        # type: () -> None

        # Although https://www.python.org/dev/peps/pep-0440 which does not allow a `-` in modern
        # versions, it also stipulates that all versions (legacy) must be handled. It turns out
        # wheel normalizes `-` to `_` and Pip has had to deal with this:
        #   https://github.com/pypa/pip/issues/1150
        #
        # We deal with this similarly.
        object.__setattr__(
            self,
            "normalized",
            cast(str, packaging_utils.canonicalize_version(self.raw)).replace("-", "_"),
        )

    @property
    def parsed_version(self):
        # type: () -> ParsedVersion
        if self._parsed_version is not None:
            return self._parsed_version

        parsed_version = packaging_version.parse(self.raw)
        object.__setattr__(self, "_parsed_version", parsed_version)
        return parsed_version

    def __str__(self):
        # type: () -> str
        return self.normalized



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).
import os.path
import shutil
import subprocess
import sys

import pytest

from pex.common import safe_rmtree
from pex.compatibility import commonpath
from pex.layout import Layout
from pex.testing import run_pex_command
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import Virtualenv

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.parametrize(
    "layout", [pytest.param(layout, id=layout.value) for layout in Layout.values()]
)
@pytest.mark.parametrize(
    "symlink_site_packages",
    [
        pytest.param(True, id="--no-venv-site-packages-copies"),
        pytest.param(False, id="--venv-site-packages-copies"),
    ],
)
def test_venv_symlink_site_packages(
    tmpdir,  # type: Any
    layout,  # type: Layout.Value
    symlink_site_packages,  # type: bool
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "pex")
    pex_root = os.path.join(str(tmpdir), "pex_root")
    venv_site_packages_copies_arg = (
        "--no-venv-site-packages-copies" if symlink_site_packages else "--venv-site-packages-copies"
    )
    result = run_pex_command(
        args=[
            "ansicolors==1.1.8",
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--venv",
            venv_site_packages_copies_arg,
            "--layout",
            layout.value,
            "--seed",
            "-o",
            pex,
        ]
    )
    result.assert_success()
    venv_pex_path = str(result.output.strip())

    safe_rmtree(pex_root)
    colors_module_realpath = os.path.realpath(
        subprocess.check_output(
            args=[sys.executable, pex, "-c", "import colors; print(colors.__file__)"]
        )
        .decode("utf-8")
        .strip()
    )

    venv_dir = os.path.dirname(venv_pex_path)
    virtualenv = Virtualenv(venv_dir)
    site_packages_dir = os.path.realpath(virtualenv.site_packages_dir)

    ansicolors_venv_package_dir_realpath = os.path.join(site_packages_dir, "colors")
    assert os.path.isdir(ansicolors_venv_package_dir_realpath)

    symlinks_expected = symlink_site_packages and layout != Layout.LOOSE
    assert os.path.islink(ansicolors_venv_package_dir_realpath) == symlinks_expected

    if symlinks_expected:
        installed_wheels_dir_realpath = os.path.realpath(os.path.join(pex_root, "installed_wheels"))
        assert installed_wheels_dir_realpath == commonpath(
            (installed_wheels_dir_realpath, colors_module_realpath)
        )
    else:
        assert (
            os.path.join(ansicolors_venv_package_dir_realpath, "__init__.py")
            == colors_module_realpath
        )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os

from pex.testing import make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_mount_respects_env(
    pex_project_dir,  # type: str
    tmpdir,  # type: Any
):
    # type: (...) -> None

    home = os.path.join(str(tmpdir), "home")

    pex_root = os.path.join(home, ".pex")
    os.makedirs(pex_root)
    os.chmod(pex_root, 0o555)
    unwritable_pex_root_warning = "PEXWarning: PEX_ROOT is configured as {}".format(pex_root)

    pex_file = os.path.join(str(tmpdir), "pex.pex")

    result = run_pex_command(
        args=[pex_project_dir, "-o", pex_file], env=make_env(HOME=home), quiet=True
    )
    result.assert_success()

    assert unwritable_pex_root_warning in result.error

    pex_root_override = os.path.join(str(tmpdir), "pex_root_override")
    result = run_pex_command(
        args=[pex_project_dir, "-o", pex_file],
        env=make_env(HOME=home, PEX_ROOT=pex_root_override),
        quiet=True,
    )
    result.assert_success()
    assert unwritable_pex_root_warning not in result.error



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import re
import subprocess
import sys

from pex.testing import PY38, PY39, PY310, ensure_python_interpreter, make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Iterable, Optional, Tuple


def test_unconstrained_universal_venv_pex(tmpdir):
    # type: (Any) -> None
    setuptools_pex = os.path.join(str(tmpdir), "setuptools.pex")
    run_pex_command(args=["setuptools==44.0.0", "-o", setuptools_pex, "--venv"]).assert_success()

    def execute_pex(
        python,  # type: str
        **extra_env  # type: str
    ):
        # type: (...) -> Tuple[bytes, bytes, int]
        process = subprocess.Popen(
            args=[
                python,
                setuptools_pex,
                "-c",
                "import sys; print('.'.join(map(str, sys.version_info[:2])))",
            ],
            env=make_env(PEX_VERBOSE=1, **extra_env),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = process.communicate()
        return stdout, stderr, process.returncode

    def assert_uses_python(
        python,  # type: str
        expected_version,  # type: Tuple[int, int]
        expected_warnings=None,  # type: Optional[Iterable[str]]
        **extra_env  # type: str
    ):
        # type: (...) -> None
        stdout, stderr, returncode = execute_pex(python, **extra_env)
        assert 0 == returncode, stderr
        assert ".".join(map(str, expected_version)) == stdout.decode("utf-8").strip()

        stderr_text = stderr.decode("utf-8").strip()
        if expected_warnings:
            assert "PEXWarning" in stderr_text
            for expected_warning in expected_warnings:
                assert re.search(expected_warning, stderr_text)
        else:
            assert "PEXWarning" not in stderr_text

    py38 = ensure_python_interpreter(PY38)
    py39 = ensure_python_interpreter(PY39)
    py310 = ensure_python_interpreter(PY310)

    assert_uses_python(python=sys.executable, expected_version=sys.version_info[:2])
    assert_uses_python(python=py38, expected_version=(3, 8))
    assert_uses_python(python=py39, expected_version=(3, 9))
    assert_uses_python(python=py310, expected_version=(3, 10))

    # When PEX_PYTHON is imprecise, the final python should be chosen by the PEX runtime.
    py39_ppp = os.path.dirname(py39)
    assert_uses_python(
        python=py310,
        expected_version=(3, 9),
        PEX_PYTHON="python3.7",
        PEX_PYTHON_PATH=py39_ppp,
        expected_warnings=[
            r"Using a venv restricted by PEX_PYTHON_PATH={ppp} for {pex} at ".format(
                ppp=py39_ppp, pex=setuptools_pex
            )
        ],
    )

    # When PEX_PYTHON is imprecise and not locked down to a minor version, a warning should be
    # issued.
    assert_uses_python(
        python=py310,
        expected_version=(3, 9),
        PEX_PYTHON="python3",
        PEX_PYTHON_PATH=py39_ppp,
        expected_warnings=[
            r"Using a venv selected by PEX_PYTHON=python3 for {pex} at".format(pex=setuptools_pex),
            r"Using a venv restricted by PEX_PYTHON_PATH={ppp} for {pex} at ".format(
                ppp=py39_ppp, pex=setuptools_pex
            ),
        ],
    )

    # When PEX_PYTHON is precise but not on PEX_PYTHON_PATH, the final python should also be chosen
    # by the PEX runtime and selection should fail.
    _, _, returncode = execute_pex(python=py310, PEX_PYTHON=py38, PEX_PYTHON_PATH=py39_ppp)
    assert 0 != returncode

    # But when PEX_PYTHON is precise and on the PEX_PYTHON_PATH, the final python should be
    # PEX_PYTHON.
    assert_uses_python(
        python=py310,
        expected_version=(3, 8),
        PEX_PYTHON=py38,
        PEX_PYTHON_PATH=os.pathsep.join(os.path.dirname(py) for py in (py38, py39)),
    )



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import filecmp
import os
import sys
from textwrap import dedent
from zipfile import ZipFile

import pytest

from pex.common import temporary_dir
from pex.compatibility import PY2
from pex.testing import (
    IS_LINUX_ARM64,
    IS_MAC_ARM64,
    IS_PYPY,
    PY27,
    PY38,
    PY310,
    PY_VER,
    create_pex_command,
    ensure_python_interpreter,
    run_command_with_jitter,
    run_commands_with_jitter,
    temporary_content,
)
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, List, Optional, Tuple


def assert_reproducible_build(
    args,  # type: List[str]
    pythons=None,  # type: Optional[Iterable[str]]
):
    # type: (...) -> None
    with temporary_dir() as td:

        def explode_pex(path):
            with ZipFile(path) as zf:
                pex_name, _ = os.path.splitext(path)
                destination_dir = os.path.join(td, "pex{}".format(pex_name))
                zf.extractall(path=destination_dir)
                return [os.path.join(destination_dir, member) for member in sorted(zf.namelist())]

        if pythons:
            pexes = run_commands_with_jitter(
                path_argument="--output-file",
                commands=[
                    create_pex_command(
                        args=args + ["--python-shebang", "/usr/bin/env python"],
                        python=python,
                        quiet=True,
                    )
                    for python in pythons
                ],
            )
        else:
            pexes = run_command_with_jitter(
                create_pex_command(args=args, quiet=True), path_argument="--output-file", count=3
            )

        pex_members = {pex: explode_pex(path=pex) for pex in pexes}
        pex1 = pexes.pop()
        for pex2 in pexes:
            # First compare file-by-file for easier debugging.
            for member1, member2 in zip(pex_members[pex1], pex_members[pex2]):
                assert not os.path.isdir(member1) ^ os.path.isdir(member2)
                if os.path.isdir(member1):
                    continue
                # Check that each file has the same content.
                with open(member1, "rb") as f1, open(member2, "rb") as f2:
                    assert list(f1.readlines()) == list(
                        f2.readlines()
                    ), "{} and {} have different content.".format(member1, member2)
                # Check that the entire file is equal, including metadata.
                assert filecmp.cmp(member1, member2, shallow=False)
            # Finally, check that the .pex files are byte-for-byte identical.
            assert filecmp.cmp(pex1, pex2, shallow=False)


@pytest.fixture(scope="module")
def major_compatible_pythons():
    # type: () -> Tuple[str, ...]
    return (
        (sys.executable, ensure_python_interpreter(PY27))
        if PY2
        else (sys.executable, ensure_python_interpreter(PY38), ensure_python_interpreter(PY310))
    )


@pytest.fixture(scope="module")
def mixed_major_pythons():
    # type: () -> Tuple[str, ...]
    return (
        sys.executable,
        ensure_python_interpreter(PY27),
        ensure_python_interpreter(PY38),
        ensure_python_interpreter(PY310),
    )


def test_reproducible_build_no_args(mixed_major_pythons):
    # type: (Tuple[str, ...]) -> None
    assert_reproducible_build([], pythons=mixed_major_pythons)


@pytest.mark.skipif(
    ((IS_MAC_ARM64 or IS_LINUX_ARM64) and PY_VER != (3, 6))
    or PY_VER > (3, 10)
    or (IS_PYPY and PY_VER > (3, 7)),
    reason=(
        "There are no pre-built binaries for the cryptography distribution for PyPy 3.8+, or "
        "for CPython 2.7 on macOS/Linux ARM64. There are also no pre-built binaries for its "
        "transitive dependency on cffi for CPython 3.11+; so this test fails for those "
        "interpreters since it requires building an sdist and that leads to an underlying C `.so`"
        "build that we have insufficient control over to make reproducible."
    ),
)
def test_reproducible_build_bdist_requirements():
    # type: () -> None
    # We test both a pure Python wheel (six) and a platform-specific wheel (cryptography).
    assert_reproducible_build(
        [
            "six==1.12.0",
            "cryptography=={version}".format(version="2.6.1" if PY2 else "3.4.8"),
        ]
    )


def test_reproducible_build_sdist_requirements(major_compatible_pythons):
    # type: (Tuple[str, ...]) -> None
    # The python-crontab sdist will be built as py2-none-any or py3-none-any depending on the
    # Python major version since it is not marked as universal in the sdist.
    assert_reproducible_build(["python-crontab==2.3.6"], pythons=major_compatible_pythons)


def test_reproducible_build_m_flag(mixed_major_pythons):
    # type: (Tuple[str, ...]) -> None
    assert_reproducible_build(["-m", "pydoc"], pythons=mixed_major_pythons)


def test_reproducible_build_c_flag_from_source(major_compatible_pythons):
    # type: (Tuple[str, ...]) -> None
    setup_cfg = dedent(
        """\
        [wheel]
        universal = 1
        """
    )
    setup_py = dedent(
        """\
        from setuptools import setup

        setup(
            name='my_app',
            entry_points={'console_scripts': ['my_app_function = my_app:do_something']},
        )
        """
    )
    my_app = dedent(
        """\
        def do_something():
            return "reproducible"
        """
    )
    with temporary_content(
        {"setup.cfg": setup_cfg, "setup.py": setup_py, "my_app.py": my_app}
    ) as project_dir:
        assert_reproducible_build(
            [project_dir, "-c", "my_app_function"],
            # Modern Pip / Setuptools produce different metadata for sdists than legacy Pip /
            # Setuptools; so we don't mix them.
            pythons=major_compatible_pythons,
        )


def test_reproducible_build_c_flag_from_dependency(major_compatible_pythons):
    # type: (Tuple[str, ...]) -> None
    # The futurize script installed depends on the version of python being used; so we don't try
    # to mix Python 2 with Python 3 as in many other reproducibility tests.
    assert_reproducible_build(
        ["future==0.17.1", "-c", "futurize"], pythons=major_compatible_pythons
    )


def test_reproducible_build_python_flag(mixed_major_pythons):
    # type: (Tuple[str, ...]) -> None
    assert_reproducible_build(
        ["--python", "python2.7", "--python-path", os.pathsep.join(mixed_major_pythons)],
        pythons=mixed_major_pythons,
    )


def test_reproducible_build_python_shebang_flag():
    # type: () -> None
    # Passing `python_versions` override `--python-shebang`; so we don't do that here.
    assert_reproducible_build(["--python-shebang=/usr/bin/python"])



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess

from pex.common import temporary_dir
from pex.testing import make_env, run_pex_command


def test_force_local_implicit_ns_packages():
    # type: () -> None
    # This was a minimal repro for the issue documented in #598.
    with temporary_dir() as out:
        tcl_pex = os.path.join(out, "tcl.pex")
        run_pex_command(["twitter.common.lang==0.3.9", "-o", tcl_pex])

        subprocess.check_call(
            [tcl_pex, "-c", "from twitter.common.lang import Singleton"],
            env=make_env(PEX_FORCE_LOCAL="1", PEX_PATH=tcl_pex),
        )



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import glob
import json
import os
import re
import subprocess
import sys
from collections import defaultdict
from contextlib import contextmanager
from textwrap import dedent

import pytest

from pex.common import chmod_plus_x, safe_mkdir, safe_mkdtemp, temporary_dir, touch
from pex.executor import Executor
from pex.interpreter import PythonInterpreter
from pex.jobs import Job
from pex.pyenv import Pyenv
from pex.testing import (
    PY38,
    PY39,
    PY310,
    PY_VER,
    all_pythons,
    ensure_python_distribution,
    ensure_python_interpreter,
    ensure_python_venv,
    environment_as,
    pushd,
)
from pex.typing import TYPE_CHECKING
from pex.variables import ENV

try:
    from unittest.mock import Mock, patch  # type: ignore[import]
except ImportError:
    from mock import Mock, patch  # type: ignore[misc,import]

if TYPE_CHECKING:
    from typing import Any, Iterator, List, Tuple

    from pex.interpreter import InterpreterOrError


def tuple_from_version(version_string):
    # type: (str) -> Tuple[int, ...]
    return tuple(int(component) for component in version_string.split("."))


class TestPythonInterpreter(object):
    def test_all_does_not_raise_with_empty_path_envvar(self):
        # type: () -> None
        """additionally, tests that the module does not raise at import."""
        with patch.dict(os.environ, clear=True), PythonInterpreter._cleared_memory_cache():
            PythonInterpreter.all()

    TEST_INTERPRETER1_VERSION = PY39
    TEST_INTERPRETER1_VERSION_TUPLE = tuple_from_version(TEST_INTERPRETER1_VERSION)

    TEST_INTERPRETER2_VERSION = PY38
    TEST_INTERPRETER2_VERSION_TUPLE = tuple_from_version(TEST_INTERPRETER2_VERSION)

    @pytest.fixture
    def test_interpreter1(self):
        # type: () -> str
        return ensure_python_interpreter(self.TEST_INTERPRETER1_VERSION)

    @pytest.fixture
    def test_interpreter2(self):
        # type: () -> str
        return ensure_python_interpreter(self.TEST_INTERPRETER2_VERSION)

    def test_interpreter_versioning(self, test_interpreter1):
        # type: (str) -> None
        py_interpreter = PythonInterpreter.from_binary(test_interpreter1)
        assert py_interpreter.identity.version == self.TEST_INTERPRETER1_VERSION_TUPLE

    def test_interpreter_caching(self, test_interpreter1, test_interpreter2):
        # type: (str, str) -> None
        py_interpreter1 = PythonInterpreter.from_binary(test_interpreter1)
        py_interpreter2 = PythonInterpreter.from_binary(test_interpreter2)
        assert py_interpreter1 is not py_interpreter2
        assert py_interpreter2.identity.version == self.TEST_INTERPRETER2_VERSION_TUPLE

        py_interpreter3 = PythonInterpreter.from_binary(test_interpreter1)
        assert py_interpreter1 is py_interpreter3

    def test_nonexistent_interpreter(self):
        # type: () -> None
        with pytest.raises(PythonInterpreter.InterpreterNotFound):
            PythonInterpreter.from_binary("/nonexistent/path")

    def test_binary_name_matching(self):
        # type: () -> None
        valid_binary_names = (
            "Python",
            "pypy",
            "pypy2",
            "pypy3",
            "pypy3.6",
            "pypy3.6m",
            "python",
            "python2",
            "python2.7",
            "python2.7m",
            "python3",
            "python3.6",
            "python3.6m",
            "python3.10",
            "python3.10m",
            "python3.99",
            "python3.99m",
            "python3.123",
            "python3.123m",
        )

        matches = PythonInterpreter._matches_binary_name
        for name in valid_binary_names:
            assert matches(name), "Expected {} to be valid binary name".format(name)

    def test_iter_interpreter_some(self, test_interpreter1, test_interpreter2):
        # type: (str, str) -> None
        assert [
            PythonInterpreter.from_binary(test_interpreter1),
            PythonInterpreter.from_binary(test_interpreter2),
        ] == list(PythonInterpreter.iter_candidates(paths=[test_interpreter1, test_interpreter2]))

    def test_iter_interpreter_none(self):
        # type: () -> None
        assert [] == list(PythonInterpreter.iter_candidates(paths=[os.devnull]))

    def test_iter_candidates_empty_paths(self, test_interpreter1):
        # type: (str) -> None
        # Whereas `paths=None` should inspect $PATH, `paths=[]` means to search nothing.
        with environment_as(PATH=test_interpreter1):
            assert [] == list(PythonInterpreter.iter_candidates(paths=[]))
            assert [PythonInterpreter.from_binary(test_interpreter1)] == list(
                PythonInterpreter.iter_candidates(paths=None)
            )

    @pytest.fixture
    def invalid_interpreter(self):
        # type: () -> Iterator[str]
        with temporary_dir() as bin_dir:
            invalid_interpreter = os.path.join(bin_dir, "python")
            touch(invalid_interpreter)
            yield invalid_interpreter

    def assert_error(self, result, expected_python):
        # type: (InterpreterOrError, str) -> None
        assert isinstance(result, tuple)
        python, error_message = result
        assert expected_python == python
        assert isinstance(error_message, str)
        assert len(error_message) > 0

    def test_iter_interpreter_errors(self, invalid_interpreter):
        # type: (str) -> None
        results = list(PythonInterpreter.iter_candidates(paths=[invalid_interpreter]))

        assert len(results) == 1
        self.assert_error(results[0], invalid_interpreter)

    def test_iter_interpreter_mixed(
        self, test_interpreter1, test_interpreter2, invalid_interpreter
    ):
        # type: (str, str, str) -> None
        results = list(
            PythonInterpreter.iter_candidates(
                paths=[test_interpreter1, invalid_interpreter, test_interpreter2]
            )
        )

        assert len(results) == 3
        assert [
            PythonInterpreter.from_binary(path) for path in (test_interpreter1, test_interpreter2)
        ] == [result for result in results if isinstance(result, PythonInterpreter)]
        errors = [result for result in results if not isinstance(result, PythonInterpreter)]
        assert len(errors) == 1
        self.assert_error(errors[0], invalid_interpreter)

    def test_iter_interpreter_path_filter(self, test_interpreter1, test_interpreter2):
        # type: (str, str) -> None
        assert [PythonInterpreter.from_binary(test_interpreter2)] == list(
            PythonInterpreter.iter_candidates(
                paths=[test_interpreter1, test_interpreter2],
                path_filter=lambda path: path == test_interpreter2,
            )
        )

    def test_iter_interpreter_path_filter_symlink(self, test_interpreter1, test_interpreter2):
        # type: (str, str) -> None
        with temporary_dir() as tmpdir:
            # PythonInterpreter._find_pyvenv_cfg looks in the binary's parent dir for a
            # pyvenv.cfg file. We use a subdir to ensure that a random such file in /tmp
            # doesn't impact this test.
            bin_dir = os.path.join(tmpdir, "bin_dir")
            os.mkdir(bin_dir)
            os.symlink(test_interpreter2, os.path.join(bin_dir, "jake"))

            # Verify path filtering happens before interpreter resolution, which os.path.realpaths
            # the interpreter binary. This supports specifying a path filter like
            # "basename is python2" where discovered interpreter binaries are symlinks to more
            # specific interpreter versions, e.g.: /usr/bin/python2 -> /usr/bin/python2.7.
            expected_interpreter = PythonInterpreter.from_binary(test_interpreter2)
            assert [expected_interpreter] == list(
                PythonInterpreter.iter_candidates(
                    paths=[test_interpreter1, bin_dir],
                    path_filter=lambda path: os.path.basename(path) == "jake",
                )
            )
            assert os.path.basename(expected_interpreter.binary) != "jake"

    def test_pyenv_shims(self, tmpdir):
        # type: (Any) -> None
        _, py38, _, run_pyenv = ensure_python_distribution(PY38)
        py310 = ensure_python_interpreter(PY310)

        pyenv_root = str(run_pyenv(["root"]).strip())
        pyenv_shims = os.path.join(pyenv_root, "shims")

        def pyenv_global(*versions):
            # type: (*str) -> None
            run_pyenv(["global"] + list(versions))

        def pyenv_local(*versions):
            # type: (*str) -> None
            run_pyenv(["local"] + list(versions))

        @contextmanager
        def pyenv_shell(*versions):
            # type: (*str) -> Iterator[None]
            with environment_as(PYENV_VERSION=":".join(versions)):
                yield

        pex_root = os.path.join(str(tmpdir), "pex_root")
        cwd = safe_mkdir(os.path.join(str(tmpdir), "home", "jake", "project"))
        with ENV.patch(PEX_ROOT=pex_root) as pex_env, environment_as(
            PYENV_ROOT=pyenv_root, PEX_PYTHON_PATH=pyenv_shims, **pex_env
        ), pyenv_shell(), pushd(cwd):
            pyenv = Pyenv.find()
            assert pyenv is not None
            assert pyenv_root == pyenv.root

            def interpreter_for_shim(shim_name):
                # type: (str) -> PythonInterpreter
                binary = os.path.join(pyenv_shims, shim_name)
                return PythonInterpreter.from_binary(binary, pyenv=pyenv)

            def assert_shim(
                shim_name,  # type: str
                expected_binary_path,  # type: str
            ):
                # type: (...) -> None
                python = interpreter_for_shim(shim_name)
                assert expected_binary_path == python.binary

            def assert_shim_inactive(shim_name):
                # type: (str) -> None
                with pytest.raises(PythonInterpreter.IdentificationError):
                    interpreter_for_shim(shim_name)

            pyenv_global(PY38, PY310)
            assert_shim("python", py38)
            assert_shim("python3", py38)
            assert_shim("python3.8", py38)
            assert_shim("python3.10", py310)

            pyenv_global(PY310, PY38)
            assert_shim("python", py310)
            assert_shim("python3", py310)
            assert_shim("python3.10", py310)
            assert_shim("python3.8", py38)

            pyenv_local(PY38)
            assert_shim("python", py38)
            assert_shim("python3", py38)
            assert_shim("python3.8", py38)
            assert_shim_inactive("python3.7")

            with pyenv_shell(PY310):
                assert_shim("python", py310)
                assert_shim("python3", py310)
                assert_shim("python3.10", py310)
                assert_shim_inactive("python3.8")

            with pyenv_shell(PY38, PY310):
                assert_shim("python", py38)
                assert_shim("python3", py38)
                assert_shim("python3.8", py38)
                assert_shim("python3.10", py310)

            # The shim pointer is now invalid since python3.7 was uninstalled and so
            # should be re-read and found invalid.
            py37_version_dir = os.path.dirname(os.path.dirname(py38))
            py37_deleted = "{}.uninstalled".format(py37_version_dir)
            os.rename(py37_version_dir, py37_deleted)
            try:
                assert_shim_inactive("python")
                assert_shim_inactive("python3")
                assert_shim_inactive("python3.8")
            finally:
                os.rename(py37_deleted, py37_version_dir)

            assert_shim("python", py38)


def test_latest_release_of_min_compatible_version():
    # type: () -> None
    def mock_interp(version):
        interp = Mock()
        interp.version = tuple(int(v) for v in version.split("."))
        return interp

    def assert_chosen(expected_version, other_version):
        expected = mock_interp(expected_version)
        other = mock_interp(other_version)
        assert (
            PythonInterpreter.latest_release_of_min_compatible_version([expected, other])
            == expected
        ), "{} was selected instead of {}".format(other_version, expected_version)

    # Note that we don't consider the interpreter name in comparisons.
    assert_chosen(expected_version="2.7.0", other_version="3.6.0")
    assert_chosen(expected_version="3.5.0", other_version="3.6.0")
    assert_chosen(expected_version="3.6.1", other_version="3.6.0")


def test_detect_pyvenv(tmpdir):
    # type: (Any) -> None
    venv = str(tmpdir)
    py38 = ensure_python_interpreter(PY38)
    real_interpreter = PythonInterpreter.from_binary(py38)
    real_interpreter.execute(["-m", "venv", venv])
    with pytest.raises(Executor.NonZeroExit):
        real_interpreter.execute(["-c", "import colors"])

    venv_bin_dir = os.path.join(venv, "bin")
    subprocess.check_call([os.path.join(venv_bin_dir, "pip"), "install", "ansicolors==1.1.8"])

    canonical_to_python = defaultdict(set)
    for python in glob.glob(os.path.join(venv_bin_dir, "python*")):
        venv_interpreter = PythonInterpreter.from_binary(python)
        canonical_to_python[venv_interpreter.binary].add(python)
        venv_interpreter.execute(["-c", "import colors"])

    assert (
        len(canonical_to_python) == 1
    ), "Expected exactly one canonical venv python, found: {}".format(canonical_to_python)
    canonical, pythons = canonical_to_python.popitem()

    real_python = os.path.realpath(py38)
    assert canonical != real_python
    assert os.path.dirname(canonical) == venv_bin_dir
    assert os.path.realpath(canonical) == real_python
    assert len(pythons) >= 2, "Expected at least two virtualenv python binaries, found: {}".format(
        pythons
    )


def check_resolve_venv(real_interpreter):
    # type: (PythonInterpreter) -> None
    tmpdir = safe_mkdtemp()

    def create_venv(
        interpreter,  # type: PythonInterpreter
        rel_path,  # type: str
    ):
        # type: (...) -> List[str]
        venv_dir = os.path.join(tmpdir, rel_path)
        interpreter.execute(["-m", "venv", venv_dir])
        return glob.glob(os.path.join(venv_dir, "bin", "python*"))

    assert not real_interpreter.is_venv
    assert real_interpreter is real_interpreter.resolve_base_interpreter()

    for index, python in enumerate(create_venv(real_interpreter, "first-level")):
        venv_interpreter = PythonInterpreter.from_binary(python)
        assert venv_interpreter.is_venv
        assert venv_interpreter != real_interpreter.binary
        assert real_interpreter == venv_interpreter.resolve_base_interpreter()

        for nested_python in create_venv(venv_interpreter, "second-level{}".format(index)):
            nested_venv_interpreter = PythonInterpreter.from_binary(nested_python)
            assert nested_venv_interpreter.is_venv
            assert nested_venv_interpreter != venv_interpreter
            assert nested_venv_interpreter != real_interpreter
            assert real_interpreter == nested_venv_interpreter.resolve_base_interpreter()


def test_resolve_venv():
    # type: () -> None
    real_interpreter = PythonInterpreter.from_binary(ensure_python_interpreter(PY38))
    check_resolve_venv(real_interpreter)


@pytest.mark.skipif(
    PY_VER < (3, 0), reason="Test requires the venv module which is not present in Python 2."
)
def test_resolve_venv_ambient():
    # type: () -> None
    ambient_real_interpreter = PythonInterpreter.get().resolve_base_interpreter()
    check_resolve_venv(ambient_real_interpreter)


def test_identify_cwd_isolation_issues_1231(tmpdir):
    # type: (Any) -> None

    python38, pip = ensure_python_venv(PY38)
    polluted_cwd = os.path.join(str(tmpdir), "dir")
    subprocess.check_call(args=[pip, "install", "--target", polluted_cwd, "pex==2.1.16"])

    pex_root = os.path.join(str(tmpdir), "pex_root")
    with pushd(polluted_cwd), ENV.patch(
        PEX_ROOT=pex_root
    ), PythonInterpreter._cleared_memory_cache():
        interp = PythonInterpreter.from_binary(python38)

    interp_info_files = {
        os.path.join(root, f)
        for root, _, files in os.walk(pex_root)
        for f in files
        if f == PythonInterpreter.INTERP_INFO_FILE
    }
    assert 1 == len(interp_info_files)
    with open(interp_info_files.pop()) as fp:
        assert interp.binary == json.load(fp)["binary"]


@pytest.fixture(scope="module")
def macos_monterey_interpeter(tmpdir_factory):
    # type: (Any) -> str
    pythonwrapper = os.path.join(str(tmpdir_factory.mktemp("bin")), "pythonwrapper")
    with open(pythonwrapper, "w") as fp:
        fp.write(
            dedent(
                """\
                #!/usr/bin/env bash
                echo >&2 "pythonwrapper[3129:20922] \\
                pythonwrapper is not supposed to be executed directly. Exiting."
                exit 0
                """
            )
        )
    chmod_plus_x(pythonwrapper)

    python = os.path.join(str(tmpdir_factory.mktemp("bin")), "python")
    os.symlink(pythonwrapper, python)
    return python


def test_issue_1494_job_error_not_identification_error(
    macos_monterey_interpeter,  # type: str
    tmpdir,  # type: Any
):
    # type: (...) -> None
    pex_root = os.path.join(str(tmpdir), "pex_root")
    with ENV.patch(PEX_ROOT=pex_root):
        spawned_job = PythonInterpreter._spawn_from_binary(macos_monterey_interpeter)
        with pytest.raises(Job.Error) as exc_info:
            spawned_job.await_result()
        exc_info.match(
            r"^Expected job to create file '{}/interpreters/[0-9a-f/]+/INTERP-INFO' "
            r"but it did not exist or could not be read: ".format(re.escape(pex_root))
        )
        exc_info.match(
            r"\n"
            r"STDERR:\n"
            r"pythonwrapper\[3129:20922\] "
            r"pythonwrapper is not supposed to be executed directly. Exiting.\n$"
        )


def test_issue_1494_iter(macos_monterey_interpeter):
    # type: (str) -> None
    assert [PythonInterpreter.get()] == list(
        PythonInterpreter.iter(paths=[sys.executable, macos_monterey_interpeter])
    )


def test_issue_1494_iter_candidates(macos_monterey_interpeter):
    # type: (str) -> None
    assert [
        PythonInterpreter.get(),
        (
            macos_monterey_interpeter,
            "pythonwrapper[3129:20922] pythonwrapper is not supposed to be executed directly. "
            "Exiting.",
        ),
    ] == list(PythonInterpreter.iter_candidates(paths=[sys.executable, macos_monterey_interpeter]))


@pytest.mark.parametrize(
    "python",
    [
        pytest.param(python, id=str(PythonInterpreter.from_binary(python).identity))
        for python in all_pythons()
    ],
)
def test_sys_path(python):
    # type: (str) -> None

    interp = PythonInterpreter.from_binary(python)
    _, stdout, _ = interp.execute(args=["-c", "import os, sys; print(os.linesep.join(sys.path))"])
    assert tuple(entry for entry in stdout.splitlines() if entry) == interp.sys_path, (
        'Its expected the sys_path matches the runtime sys.path with the exception of the PWD ("") '
        "head entry."
    )



# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from docutils import nodes
from docutils.parsers.rst import Directive
from sphinx import addnodes

from pex.variables import DefaultedProperty, Variables


class Vars(Directive):
    def run(self):
        def make_nodes(var_name):
            var_obj = Variables.__dict__[var_name]
            if isinstance(var_obj, DefaultedProperty):
                desc_str = var_obj._func.__doc__
            else:
                desc_str = var_obj.__doc__
            desc_str = desc_str or "NO DESC"

            sig = addnodes.desc()
            sig.append(nodes.target("", "", ids=[var_name]))
            sig.append(addnodes.desc_signature(var_name, var_name))
            desc = nodes.paragraph()
            for line in desc_str.split("\n"):
                desc += nodes.line(line, line)
            sig["objtype"] = sig["desctype"] = "var"
            return sig, desc

        return [
            node for var in dir(Variables) if var.startswith("PEX_") for node in make_nodes(var)
        ]


def setup(app):
    app.add_directive("vars", Vars)

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import sys
from textwrap import dedent

import pytest

from pex import requirements
from pex.cli.testing import run_pex3
from pex.compatibility import to_unicode
from pex.dist_metadata import Requirement
from pex.requirements import LocalProjectRequirement, Source
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


@pytest.fixture
def lockfile():
    # type: () -> str
    return os.path.join(os.path.dirname(__file__), "test_export_subset.lock.json")


def test_full(
    tmpdir,  # type: Any
    lockfile,  # type: str
):
    # type: (...) -> None

    result = run_pex3("lock", "export", lockfile)
    result.assert_success()
    export = result.output

    result = run_pex3("lock", "export-subset", "--lock", lockfile)
    result.assert_success()
    export_subset = result.output

    assert (
        export == export_subset
    ), "An export-subset with no requirements should export all requirements, just like export."

    actual_requirements = []
    for parsed_requirement in requirements.parse_requirements(Source.from_text(export_subset)):
        assert not isinstance(parsed_requirement, LocalProjectRequirement)
        actual_requirements.append(parsed_requirement.requirement)

    expected_requirements = [
        Requirement.parse(to_unicode("django==1.11.29")),
        Requirement.parse(to_unicode("argon2-cffi==20.1.0")),
        Requirement.parse(to_unicode("bcrypt==3.1.7")),
        Requirement.parse(to_unicode("pytz==2023.3")),
        Requirement.parse(to_unicode("cffi==1.15.1")),
        Requirement.parse(to_unicode("six==1.16.0")),
        Requirement.parse(to_unicode("pycparser==2.21")),
    ]
    if sys.version_info[0] == 2:
        expected_requirements.append(Requirement.parse(to_unicode("enum34==1.1.10")))

    assert sorted(expected_requirements, key=str) == sorted(actual_requirements, key=str)


def test_subset(
    tmpdir,  # type: Any
    lockfile,  # type: str
):
    # type: (...) -> None

    result = run_pex3("lock", "export-subset", "django", "--lock", lockfile)
    result.assert_success()

    assert (
        dedent(
            """\
            django==1.11.29 \\
              --hash=sha256:014e3392058d94f40569206a24523ce254d55ad2f9f46c6550b0fe2e4f94cf3f \\
              --hash=sha256:4200aefb6678019a0acf0005cd14cfce3a5e6b9b90d06145fcdd2e474ad4329c
            pytz==2023.3 \\
              --hash=sha256:a151b3abb88eda1d4e34a9814df37de2a80e301e68ba0fd856fb9b46bfbbbffb \\
              --hash=sha256:1d8ce29db189191fb55338ee6d0387d82ab59f3d00eac103412d64e0ebd0c588
            """
        )
        == result.output
    ), (
        "Django without extras only depends on pytz and both are available in the lock in sdist "
        "and universal wheel formats"
    )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import shutil
import subprocess
from textwrap import dedent

from pex.common import safe_open
from pex.testing import PY310, ensure_python_distribution, make_project, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_excepthook_scrubbing(tmpdir):
    # type: (Any) -> None

    original_python_installation, original_python, _, _ = ensure_python_distribution(PY310)

    custom_python_installation = os.path.join(str(tmpdir), "custom")
    shutil.copytree(original_python_installation, custom_python_installation)
    custom_python = os.path.join(custom_python_installation, "bin", "python")

    project_dir = os.path.join(str(tmpdir), "custom_excepthook")
    with safe_open(os.path.join(project_dir, "custom_excepthook.py"), "w") as fp:
        fp.write("def custom_excepthook(typ, val, _tb): print('EXC: {} {}'.format(typ, val))")
    with safe_open(os.path.join(project_dir, "custom_excepthook.pth"), "w") as fp:
        fp.write("import custom_excepthook; sys.excepthook = custom_excepthook.custom_excepthook")

    with safe_open(os.path.join(project_dir, "setup.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                from distutils import sysconfig

                from setuptools import setup


                setup(
                    name="custom_excepthook",
                    version="0.0.1",
                    py_modules=["custom_excepthook"],
                    data_files=[(sysconfig.get_python_lib(), ["custom_excepthook.pth"])]
                )
                """
            )
        )

    subprocess.check_call(args=[custom_python, "-m", "pip", "install", project_dir])

    src = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(src, "app.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import sys


                if getattr(sys.excepthook, "__name__", None) == "custom_excepthook":
                    import custom_excepthook

                print("SUCCESS")
                """
            )
        )

    pex = os.path.join(str(tmpdir), "pex")
    create_pex_args = ["-D", src, "-m", "app", "-o", pex]

    # Ensure this complicated test setup reproduces the original error in
    # https://github.com/pantsbuild/pex/issues/1809
    # (via https://github.com/pantsbuild/pants/issues/15877).
    run_pex_command(args=["pex==2.1.92", "-c", "pex", "--"] + create_pex_args).assert_success()

    # A Python installation without the custom excepthook installed via .pth should work just fine.
    assert "SUCCESS\n" == subprocess.check_output(args=[original_python, pex]).decode("utf-8")

    # But with the custom excepthook installed pre-PEX boot, we should reproduce earlier failures.
    process = subprocess.Popen(args=[custom_python, pex], stdout=subprocess.PIPE)
    stdout, _ = process.communicate()
    assert 0 != process.returncode
    assert (
        "EXC: <class 'ModuleNotFoundError'> No module named 'custom_excepthook'\n"
        == stdout.decode("utf-8")
    )

    # Now demonstrate the fix.
    run_pex_command(args=create_pex_args).assert_success()
    assert "SUCCESS\n" == subprocess.check_output(args=[custom_python, pex]).decode("utf-8")
